from pyravendb.raven_operations.query_operation import QueryOperation
from pyravendb.custom_exceptions.exceptions import *
from pyravendb.commands.raven_commands import GetFacetsCommand
from pyravendb.data.query import IndexQuery, QueryOperator, OrderingType, FacetQuery
from pyravendb.tools.utils import Utils
from datetime import timedelta
import time
import re


class _Token:
    def __init__(self, field_name="", value=None, token=None, write=None, **kwargs):
        self.__dict__.update({"field_name": field_name, "value": value, "token": token, "write": write, **kwargs})


class Query(object):
    where_operators = {'equals': 'equals', 'greater_than': 'greater_than',
                       'greater_than_or_equal': 'greater_than_or_equal', 'less_than': 'less_than',
                       'less_than_or_equal': 'less_than_or_equal', 'in': 'in', 'all_in': 'all_in',
                       'between': 'between', 'search': 'search', 'lucene': 'lucene', 'startsWith': 'startsWith',
                       'endsWith': 'endsWith', 'exists': 'exists'}

    rql_keyword = ("AS", "SELECT", "WHERE", "LOAD", "GROUP", "ORDER", "INCLUDE")

    def __init__(self, session):
        """
        These argument will be initialized when class is called (see __call__)
        for be able to query with the same instance of the Query class
        """

        self.session = session

        self.negate = None
        self.fields_to_fetch = None
        self._select_tokens = None
        self._from_token = None
        self._group_by_tokens = None
        self._order_by_tokens = None
        self._where_tokens = None
        self.includes = None
        self._current_clause_depth = None
        self.is_intersect = None
        self.the_wait_for_non_stale_results = False
        self.fetch = None
        self._query = None
        self.start = None
        self.page_size = None
        self.last_equality = None
        self.query_parameters = None
        self.is_distinct = None

    @property
    def not_(self):
        self._negate_next()
        return self

    def __call__(self, object_type=None, index_name=None, collection_name=None, is_map_reduce=False,
                 with_statistics=False, metadata_only=False, default_operator=None, wait_for_non_stale_results=False,
                 timeout=None, nested_object_types=None):
        """
        @param Type object_type: The type of the object we want to track the entity too.
        @param str index_name: The index name we want to apply.
        @param str collection_name: Name of the collection (mutually exclusive with indexName).
        @param bool is_map_reduce: Whether we are querying a map/reduce index(modify how we treat identifier properties)
        @param bool with_statistics: Make it True to get the query statistics as well.
        @param QueryOperator default_operator: The default query operator (OR or AND).
        @param bool wait_for_non_stale_results: Instructs the query to wait for non stale results.
        @param timedelta timeout: The time to wait for non stale result.
        @param nested_object_types: A dict of classes for nested object the key will be the name of the class and the.
        value will be the object we want to get for that attribute.
        """
        self.object_type = object_type
        self.index_name, self.collection_name = self._process_query_parameters(index_name, collection_name)
        self.nested_object_types = nested_object_types
        self.is_map_reduce = is_map_reduce
        self._with_statistics = with_statistics
        self.default_operator = default_operator
        self.metadata_only = metadata_only
        self.query_parameters = {}
        self.negate = False
        self.fields_to_fetch = None
        self._select_tokens = []  # List[_Token]
        self._from_token = None
        self._group_by_tokens = []  # List[_Token]
        self._order_by_tokens = []  # List[_Token]
        self._where_tokens = []  # List[_Token]
        self.includes = set()
        self._current_clause_depth = 0
        self.is_intersect = False
        self.start = 0
        self.page_size = None
        self.cutoff_etag = None
        self.wait_for_non_stale_results = wait_for_non_stale_results
        self.timeout = timeout if timeout is not None else self.session.conventions.timeout
        self.fetch = None
        self._query = None
        self.last_equality = None
        self.is_distinct = False

        return self

    def __iter__(self):
        return iter(self._execute_query())

    def __str__(self):
        return self._build_query()

    def _process_query_parameters(self, index_name, collection_name):
        if index_name and collection_name:
            raise InvalidOperationException(
                "Parameters 'index_name' and 'collection_name' are mutually exclusive. "
                "Please specify only one of them.")

        if not index_name and not collection_name:
            if not self.object_type:
                collection_name = "@all_docs"
            else:
                collection_name = self.session.conventions.default_transform_plural(self.object_type.__name__)

        return "dynamic" if index_name is None else index_name, collection_name

    def _build_query(self):
        if self._query:
            return self._query

        if self._current_clause_depth != 0:
            raise InvalidOperationException(
                "A clause was not closed correctly within this query, current clause depth = {0}".format(
                    self._current_clause_depth))

        query_builder = []
        self.build_from(query_builder)
        self.build_group_by(query_builder)
        self.build_where(query_builder)
        self.build_order_by(query_builder)
        self.build_select(query_builder)
        self.build_include(query_builder)

        return "".join(query_builder)

    def build_from(self, query_builder):
        if not self.index_name and not self.collection_name:
            raise NotSupportedException("Either index_name or collection_name must be specified")

        if self.index_name == "dynamic":
            query_builder.append("FROM ")
            if Utils.contains_any(self.collection_name, [' ', '\t', '\r', '\n', '\v']):
                if '"' in self.collection_name:
                    raise ValueError("Collection name cannot contain a quote, but was: " + self.collection_name)
                query_builder.append('"' + self.collection_name + '"')
            else:
                query_builder.append(self.collection_name)
        else:
            query_builder.append("FROM INDEX '" + self.index_name + "'")

    def build_group_by(self, query_builder):
        if len(self._group_by_tokens) > 0:
            query_builder.append(" GROUP BY ")
            index = 0
            for token in self._group_by_tokens:
                if index > 0:
                    query_builder.append(",")
                query_builder.append(token.write)
                index += 1

    def build_where(self, query_builder):
        if len(self._where_tokens) > 0:
            query_builder.append(" WHERE ")
            if self.is_intersect:
                self.query_builder.append("intersect(")

            for token in self._where_tokens:
                Query._add_space_if_needed(query_builder)
                query_builder.append(token.write)

            if self.is_intersect:
                query_builder.append(") ")

    def build_order_by(self, query_builder):
        if len(self._order_by_tokens) > 0:
            query_builder.append(" ORDER BY ")
            first = True
            for token in self._order_by_tokens:
                if not first and self._order_by_tokens[-1] is not None:
                    query_builder.append(", ")
                first = False
                query_builder.append(token.write)
                if token.ordering != OrderingType.str:
                    query_builder.append(token.ordering)
                if token.descending:
                    query_builder.append(" DESC")

    def build_select(self, query_builder):
        select_token_length = len(self._select_tokens)
        if select_token_length > 0:
            index = 0
            query_builder.append(" SELECT ")
            if select_token_length == 1 and self._select_tokens[0].token == "distinct":
                query_builder.append("DISTINCT *")
                return

            for token in self._select_tokens:
                if index > 0 and self._select_tokens[index - 1].token != "distinct":
                    query_builder.append(",")

                Query._add_space_if_needed(query_builder)
                query_builder.append(token.write)
                index += 1

    def build_include(self, query_builder):
        if self.includes is not None and len(self.includes) > 0:
            query_builder.append(" INCLUDE ")
            first = True
            for include in self.includes:
                if not first:
                    query_builder.append(",")
                first = False
                required_quotes = False
                if not re.match("^[A-Za-z0-9_]+$", include):
                    required_quotes = True
                if required_quotes:
                    query_builder.append("'{0}'".format(include.replace("'", "\\'")))
                else:
                    query_builder.append(include)

    def assert_no_raw_query(self):
        if self._query:
            raise InvalidOperationException(
                "raw_query was called, cannot modify this query by calling on operations "
                "that would modify the query (such as where, select, order_by, group_by, etc)")

    @staticmethod
    def _get_rql_write_case(token):
        write = None
        if token.token == "in":
            write = " IN ($" + token.value + ")"
        elif token.token == "all_in":
            write = " ALL IN ($" + token.value + ")"
        elif token.token == "between":
            write = "".join([" BETWEEN $", str(token.value[0]), " AND $", str(token.value[1])])
        elif token.token == "equals":
            write = " = $" + token.value
        elif token.token == "not_equals":
            write = " != $" + token.value
        elif token.token == "greater_than":
            write = " > $" + token.value
        elif token.token == "greater_than_or_equal":
            write = " >= $" + token.value
        elif token.token == "less_than":
            write = " < $" + token.value
        elif token.token == "less_than_or_equal":
            write = " <= $" + token.value
        elif token.token == "search":
            write_builder = [", $", token.value]
            search_operator = getattr(token, "search_operator", None)
            if search_operator and search_operator == QueryOperator.AND:
                write_builder.append(", AND")
            write_builder.append(")")
            write = "".join(write_builder)
        elif token.token == "lucene":
            write = ", $" + token.value + ")"
        elif token.token == "startsWith":
            write = ", $" + token.value + ")"
        elif token.token == "endsWith":
            write = ", $" + token.value + ")"
        elif token.token == "exists":
            write = ")"
        elif token.token == "select":
            write_builder = []
            for value in token.value:
                if len(write_builder) > 0:
                    write_builder.append(", ")
                    if value.upper() in Query.rql_keyword:
                        value = "'{0}'".format(value)
                write_builder.append(value)
            write = "".join(write_builder)

        if write is None:
            raise AttributeError("token.token don't match any of the cases for rql builder")
        return write

    @staticmethod
    def rql_where_write(token):
        where_rql = ""
        boost = getattr(token, "boost", None)
        if boost:
            where_rql += "boost("
        fuzzy = getattr(token, "fuzzy", None)
        if fuzzy:
            where_rql += "fuzzy("
        proximity = getattr(token, "proximity", None)
        if proximity:
            where_rql += "proximity("
        exact = getattr(token, "exact", False)
        if exact:
            where_rql += "exact("

        if token.token in ["search", "lucene", "startsWith", "endsWith", "exists"]:
            where_rql += token.token + "("

        where_rql += token.field_name + Query._get_rql_write_case(token)

        if exact:
            where_rql += ")"
        if proximity:
            where_rql += "," + proximity + ")"
        if fuzzy:
            where_rql += "," + fuzzy + ")"
        if boost:
            where_rql += "," + str(boost) + ")"

        return where_rql

    @staticmethod
    def _add_space_if_needed(query_builder):
        if len(query_builder) > 0:
            if not query_builder[-1].endswith(("(", ")", ",", " ")):
                query_builder.append(" ")

    @staticmethod
    def escape_if_needed(name):
        if name:
            escape = False
            first = True
            for c in name:
                if first:
                    if not c.isalpha() and c != '_' and c != '@':
                        escape = True
                        break
                    first = False
                else:
                    if (not c.isalpha() and not c.isdigit()) and c != '_' and c != '@' and c != '[' and c != ']':
                        escape = True
                        break

            if escape or name.upper() in Query.rql_keyword:
                return "'{0}'".format(name)

        return name

    def negate_if_needed(self, field_name):
        if self.negate:
            self.negate = False

            if len(self._where_tokens) == 0:
                if field_name is not None:
                    self.where_exists(field_name)
                else:
                    self.where_true()

                self.and_also()
            self._where_tokens.append(_Token(write="NOT"))

    def add_query_parameter(self, value):
        parameter_name = "p{0}".format(len(self.query_parameters))
        self.query_parameters[parameter_name] = value
        return parameter_name

    def _add_operator_if_needed(self):
        self.assert_no_raw_query()

        if len(self._where_tokens) > 0:
            query_operator = None
            last_token = self._where_tokens[-1]
            if last_token is not None and last_token.token in Query.where_operators:
                query_operator = QueryOperator.AND if self.default_operator == QueryOperator.AND else QueryOperator.OR
            search_operator = getattr(last_token, "search_operator", None)
            if search_operator and search_operator != QueryOperator.OR:
                # default to OR operator after search if AND was not specified explicitly
                query_operator = QueryOperator.OR

            if query_operator:
                self._where_tokens.append(_Token(write=str(query_operator)))

    def intersect(self):
        if len(self._where_tokens) > 0:
            last = self._where_tokens[-1]
            if last.token in Query.where_operators:
                self.is_intersect = True
                self._where_tokens.append(_Token(value=None, token="intersect", write=","))
                return self
            else:
                raise InvalidOperationException("Cannot add INTERSECT at this point.")

    def raw_query(self, query, query_parameters=None):
        """
        To get all the document that equal to the query

        @param str query: The rql query
        @param dict query_parameters: Add query parameters to the query {key : value}
        """
        self.assert_no_raw_query()

        if len(self._where_tokens) != 0 or len(self._select_tokens) != 0 or len(
                self._order_by_tokens) != 0 or len(self._group_by_tokens) != 0:
            raise InvalidOperationException(
                "You can only use raw_query on a new query, without applying any operations "
                "(such as where, select, order_by, group_by, etc)")

        if query_parameters:
            self.query_parameters = query_parameters
        self._query = query
        return self

    def where_equals(self, field_name, value, exact=False):
        """
        To get all the document that equal to the value in the given field_name

        @param str field_name: The field name in the index you want to query.
        @param value: The value will be the fields value you want to query
        @param bool exact: If True getting exact match of the query
        """
        if field_name is None:
            raise ValueError("None field_name is invalid")

        field_name = Query.escape_if_needed(field_name)
        self._add_operator_if_needed()

        token = "equals"
        if self.negate:
            self.negate = False
            token = "not_equals"

        self.last_equality = {field_name: value}
        token = _Token(field_name=field_name, value=self.add_query_parameter(value), token=token, exact=exact)
        token.write = self.rql_where_write(token)
        self._where_tokens.append(token)

        return self

    def where_not_equals(self, field_name, value, exact=False):
        if not self.negate:
            self.negate = True
        return self.where_equals(field_name, value, exact)

    def where_exists(self, field_name):
        field_name = Query.escape_if_needed(field_name)

        self._add_operator_if_needed()
        self.negate_if_needed(field_name)
        self._where_tokens.append(_Token(field_name=field_name, value=None, token="exists", write=")"))
        return self

    def where_true(self):
        self._add_operator_if_needed()
        self.negate_if_needed(None)

        self._where_tokens.append(_Token(token="true_token", write="true"))
        return self

    def where(self, exact=False, **kwargs):
        """
        To get all the document that equal to the value within kwargs with the specific key

        @param bool exact: If True getting exact match of the query
        @param kwargs: the keys of the kwargs will be the fields name in the index you want to query.
        The value will be the the fields value you want to query
        (if kwargs[field_name] is a list it will behave has the where_in method)
        """
        for field_name in kwargs:
            if isinstance(kwargs[field_name], list):
                self.where_in(field_name, kwargs[field_name], exact)
            else:
                self.where_equals(field_name, kwargs[field_name], exact)
        return self

    def search(self, field_name, search_terms, operator=QueryOperator.OR):
        """
        For more complex text searching

        @param str field_name: The field name in the index you want to query.
        :type str
        @param str search_terms: The terms you want to query
        @param QueryOperator operator: OR or AND
        """

        if field_name is None:
            raise ValueError("None field_name is invalid")

        field_name = Query.escape_if_needed(field_name)
        self._add_operator_if_needed()
        self.negate_if_needed(field_name)

        self.last_equality = {field_name: "(" + search_terms + ")" if ' ' in search_terms else search_terms}
        token = _Token(field_name=field_name, token="search", value=self.add_query_parameter(search_terms),
                       search_operator=operator)
        token.write = self.rql_where_write(token)
        self._where_tokens.append(token)
        return self

    def where_ends_with(self, field_name, value):
        """
        To get all the document that ends with the value in the giving field_name

        @param str field_name:The field name in the index you want to query.
        @param str value: The value will be the fields value you want to query
        """
        if field_name is None:
            raise ValueError("None field_name is invalid")

        field_name = Query.escape_if_needed(field_name)
        self._add_operator_if_needed()
        self.negate_if_needed(field_name)

        self.last_equality = {field_name: value}
        token = _Token(field_name=field_name, token="endsWith", value=self.add_query_parameter(value))
        token.write = self.rql_where_write(token)
        self._where_tokens.append(token)

        return self

    def where_starts_with(self, field_name, value):
        """
        To get all the document that starts with the value in the giving field_name

        @param str field_name:The field name in the index you want to query.
        @param str value: The value will be the fields value you want to query
        """
        if field_name is None:
            raise ValueError("None field_name is invalid")

        field_name = Query.escape_if_needed(field_name)
        self._add_operator_if_needed()
        self.negate_if_needed(field_name)

        self.last_equality = {field_name: value}
        token = _Token(field_name=field_name, token="startsWith", value=self.add_query_parameter(value))
        token.write = self.rql_where_write(token)
        self._where_tokens.append(token)

        return self

    def where_in(self, field_name, values, exact=False):
        """
        Check that the field has one of the specified values

        @param str field_name: Name of the field
        @param str values: The values we wish to query
        @param bool exact: Getting the exact query (ex. case sensitive)
        """
        field_name = Query.escape_if_needed(field_name)
        self._add_operator_if_needed()
        self.negate_if_needed(field_name)

        token = _Token(field_name=field_name, value=self.add_query_parameter(list(Utils.unpack_iterable(values))),
                       token="in", exact=exact)
        token.write = self.rql_where_write(token)
        self._where_tokens.append(token)

        return self

    def where_between(self, field_name, start, end, exact=False):
        field_name = Query.escape_if_needed(field_name)

        self._add_operator_if_needed()
        self.negate_if_needed(field_name)

        if isinstance(start, timedelta):
            start = Utils.timedelta_tick(start)
        if isinstance(end, timedelta):
            end = Utils.timedelta_tick(end)

        from_parameter_name = self.add_query_parameter("*" if start is None else start)
        to_parameter_name = self.add_query_parameter("NULL" if end is None else end)

        token = _Token(field_name=field_name, token="between", value=(from_parameter_name, to_parameter_name),
                       exact=exact)
        token.write = self.rql_where_write(token)
        self._where_tokens.append(token)

        return self

    def where_greater_than(self, field_name, value):
        field_name = Query.escape_if_needed(field_name)
        self._add_operator_if_needed()
        self.negate_if_needed(field_name)

        if isinstance(value, timedelta):
            value = Utils.timedelta_tick(value)

        token = _Token(field_name=field_name, token="greater_than", value=self.add_query_parameter(value))
        token.write = self.rql_where_write(token)
        self._where_tokens.append(token)

        return self

    def where_greater_than_or_equal(self, field_name, value):
        field_name = Query.escape_if_needed(field_name)
        self._add_operator_if_needed()
        self.negate_if_needed(field_name)

        if isinstance(value, timedelta):
            value = Utils.timedelta_tick(value)

        token = _Token(field_name=field_name, token="greater_than_or_equal", value=self.add_query_parameter(value))
        token.write = self.rql_where_write(token)
        self._where_tokens.append(token)

        return self

    def where_less_than(self, field_name, value):
        field_name = Query.escape_if_needed(field_name)
        self._add_operator_if_needed()
        self.negate_if_needed(field_name)

        if isinstance(value, timedelta):
            value = Utils.timedelta_tick(value)

        token = _Token(field_name=field_name, token="less_than", value=self.add_query_parameter(value))
        token.write = self.rql_where_write(token)
        self._where_tokens.append(token)

        return self

    def where_less_than_or_equal(self, field_name, value):
        field_name = Query.escape_if_needed(field_name)
        self._add_operator_if_needed()
        self.negate_if_needed(field_name)

        if isinstance(value, timedelta):
            value = Utils.timedelta_tick(value)

        token = _Token(field_name=field_name, token="less_than_or_equal", value=self.add_query_parameter(value))
        token.write = self.rql_where_write(token)
        self._where_tokens.append(token)

        return self

    def where_not_none(self, field_name):
        self.and_also().not_.where_equals(field_name, None)
        return self

    def add_order(self, field_name, descending, ordering):
        """
        @param str field_name: The field you want to order
        @param bool descending: In descending Order
        @param OrderingType ordering: The field_name type (str, long, float or alpha_numeric)
        :return:
        """
        return self.order_by(field_name, ordering, descending=descending)

    def order_by(self, field_name, ordering=OrderingType.str, descending=False):
        self.assert_no_raw_query()
        field_name = Query.escape_if_needed(field_name)
        self._order_by_tokens.append(
            _Token(field_name=field_name, token="order_by", write=field_name, descending=descending,
                   ordering=ordering))

        return self

    def order_by_descending(self, field_name, ordering=OrderingType.str):
        return self.order_by(field_name, ordering, descending=True)

    def order_by_score(self):
        return self.order_by("score()")

    def order_by_score_descending(self):
        return self.order_by("score()", descending=True)

    def random_ordering(self):
        return self.order_by("random()")

    def and_also(self):
        if len(self._where_tokens) > 0:
            last = self._where_tokens[-1]
            if last:
                if isinstance(self.query_parameters[last.value], QueryOperator):
                    raise InvalidOperationException("Cannot add AND, previous token was already an QueryOperator.")

                self._where_tokens.append(_Token(value=QueryOperator.AND, write="AND"))

        return self

    def distinct(self):
        if self.is_distinct:
            raise InvalidOperationException("This is already a distinct query.")
        self.is_distinct = True
        self._select_tokens.insert(0, _Token(token="distinct"))
        return self

    def select(self, *args):
        """
        @param str args: The fields to fetch
        """

        if len(args) > 0:
            self.fields_to_fetch = args
            if len(self._select_tokens) == 0:
                token = _Token(token="select", value=args)
                token.write = self._get_rql_write_case(token)
                self._select_tokens.append(token)
            else:
                replaced = False
                for token in self._select_tokens:
                    if token.token == "select":
                        token.value = args
                        token.write = self._get_rql_write_case(token)
                        replaced = True
                        break
                if not replaced:
                    token = _Token(token="select", value=args)
                    token.write = self._get_rql_write_case(token)
                    self._select_tokens.append(token)
        return self

    def or_else(self):
        if len(self.query_builder) > 0:
            self.query_builder += " OR"
        return self

    def boost(self, boost):
        if boost != 1:
            try:
                last = self._where_tokens[-1]
            except IndexError:
                raise InvalidOperationException("Missing where clause")
            if boost <= 0:
                raise ArgumentOutOfRangeException("boost", "Boost factor must be a positive number")

            setattr(last, "boost", boost)
            last.write = self.rql_where_write(last)
            return self

    def take(self, count):
        self.page_size = count
        return self

    def skip(self, count):
        self.start = count
        return self

    def _negate_next(self):
        """
        Negate the next operation
        """

        self.negate = not self.negate

    def include(self, path):
        self.includes.add(path)
        return self

    def get_last_equality_term(self):
        """
        The last term that we asked the query to use equals on
        """
        return self.last_equality

    def to_facets(self, facets, start=0, page_size=None):
        """
        Query the facets results for this query using the specified list of facets with the given start and pageSize

        @param List[Facet] facets: List of facets
        @param int start:  Start index for paging
        @param page_size: Paging PageSize. If set, overrides Facet.max_result
        """

        if len(facets) == 0:
            raise ValueError("Facets must contain at least one entry", "facets")
        str_query = self.__str__()
        facet_query = FacetQuery(str_query, None, facets, start, page_size, query_parameters=self.query_parameters,
                                 wait_for_non_stale_results=self.wait_for_non_stale_results,
                                 wait_for_non_stale_results_timeout=self.timeout, cutoff_etag=self.cutoff_etag)

        command = GetFacetsCommand(query=facet_query)
        return self.session.requests_executor.execute(command)

    def get_index_query(self):
        return IndexQuery(query=self.__str__(), query_parameters=self.query_parameters, start=self.start,
                          page_size=self.page_size, cutoff_etag=self.cutoff_etag,
                          wait_for_non_stale_results=self.wait_for_non_stale_results,
                          wait_for_non_stale_results_timeout=self.timeout)

    def _execute_query(self):
        conventions = self.session.conventions
        end_time = time.time() + self.timeout.seconds
        query = self._build_query()
        while True:
            index_query = IndexQuery(query=query, query_parameters=self.query_parameters, start=self.start,
                                     page_size=self.page_size, cutoff_etag=self.cutoff_etag,
                                     wait_for_non_stale_results=self.wait_for_non_stale_results,
                                     wait_for_non_stale_results_timeout=self.timeout)

            query_command = QueryOperation(session=self.session, index_name=self.index_name,
                                           index_query=index_query,
                                           metadata_only=self.metadata_only).create_request()
            response = self.session.requests_executor.execute(query_command)
            if response is None:
                return []

            if response["IsStale"] and self.wait_for_non_stale_results:
                if time.time() > end_time:
                    raise ErrorResponseException("The index is still stale after reached the timeout")
                continue
            break

        results = []
        response_results = response.pop("Results")
        response_includes = response.pop("Includes", None)
        self.session.save_includes(response_includes)
        for result in response_results:
            entity, metadata, original_metadata = Utils.convert_to_entity(result, self.object_type, conventions,
                                                                          self.nested_object_types,
                                                                          fetch=False if not self.fetch else True)
            if self.object_type != dict and not self.fields_to_fetch:
                self.session.save_entity(key=original_metadata.get("@id", None), entity=entity,
                                         original_metadata=original_metadata,
                                         metadata=metadata, document=result)
            results.append(entity)

        if self._with_statistics:
            return results, response
        return results



from pyravendb.tests.test_base import TestBase
from pyravendb.data.query import Facet, FacetMode
from pyravendb.raven_operations.maintenance_operations import PutIndexesOperation
from pyravendb.data.indexes import IndexDefinition
import unittest


class Product(object):
    def __init__(self, name, price_per_unit):
        self.name = name
        self.price_per_unit = price_per_unit


class ProductsAndPricePerUnit:
    def __init__(self):
        self.maps = ("from product in docs.Products "
                     "select new {"
                     "price_per_unit = product.price_per_unit}")
        self.index_definition = IndexDefinition(name=ProductsAndPricePerUnit.__name__, maps=self.maps)

    def execute(self, store):
        store.maintenance.send(PutIndexesOperation(self.index_definition))


class TestFacets(TestBase):
    def setUp(self):
        super(TestFacets, self).setUp()
        with self.store.open_session() as session:
            session.store(Product("TV", 1022))
            session.store(Product("jacket", 100))
            session.save_changes()

        ProductsAndPricePerUnit().execute(self.store)
        self.facets = [Facet("Products", mode=FacetMode.ranges),
                       Facet("price_per_unit_L_Range", ranges=["{100 TO 3000]", "[402 TO 2000]"], mode=FacetMode.ranges)]

    def tearDown(self):
        super(TestFacets, self).tearDown()
        self.delete_all_topology_files()

    def test_facets_with_documents(self):
        with self.store.open_session() as session:
            query_results = session.query(object_type=Product,
                                          index_name=ProductsAndPricePerUnit.__name__).where_greater_than(
                "price_per_unit", 99).to_facets(self.facets)
            assert len(query_results['Results']) > 0

if __name__ == "__main__":
    unittest.main()



class InvalidOperationException(Exception):
    pass


class ErrorResponseException(Exception):
    pass


class DocumentDoesNotExistsException(Exception):
    pass


class NonUniqueObjectException(Exception):
    pass


class FetchConcurrencyException(Exception):
    pass


class ArgumentOutOfRangeException(Exception):
    pass


class DatabaseDoesNotExistException(Exception):
    pass


class AuthorizationException(Exception):
    pass


class IndexDoesNotExistException(Exception):
    pass


class TimeoutException(Exception):
    pass


class AuthenticationException(Exception):
    pass


class AllTopologyNodesDownException(Exception):
    pass


class UnsuccessfulRequestException(Exception):
    pass


class AggregateException(Exception):
    pass


class NotSupportedException(Exception):
    pass


class AggregateException(Exception):
    pass


class ChangeProcessingException(Exception):
    pass


class ChangeProcessingException(Exception):
    pass


# <---------- Subscription Exceptions ---------->

class SubscriptionException(Exception):
    pass


class SubscriptionInUseException(SubscriptionException):
    pass


class SubscriptionClosedException(SubscriptionException):
    pass


class SubscriptionInvalidStateException(SubscriptionException):
    pass


class SubscriptionDoesNotExistException(SubscriptionException):
    pass


class SubscriptionDoesNotBelongToNodeException(SubscriptionException):
    def __init__(self, appropriate_node, message):
        super(SubscriptionDoesNotBelongToNodeException, self).__init__(message)
        self.appropriate_node = appropriate_node


class SubscriptionChangeVectorUpdateConcurrencyException(SubscriptionException):
    pass


class SubscriberErrorException(Exception):
    pass



from pyravendb.custom_exceptions.exceptions import InvalidOperationException
from threading import Lock, Thread
from pyravendb.tools.utils import Utils


class PropagatingThread(Thread):
    def run(self):
        self.exception = None
        try:
            self.ret = self._target(*self._args, **self._kwargs)
        except BaseException as e:
            self.exception = e

    def join(self):
        super(PropagatingThread, self).join()
        if self.exception:
            raise self.exception
        return self.ret


class ServerNode(object):
    def __init__(self, url, database=None, cluster_tag=None):
        self.url = url
        self.database = database
        self.cluster_tag = cluster_tag
        self._response_time = []
        self._is_rate_surpassed = None

    @property
    def response_time(self):
        return self._response_time

    @response_time.setter
    def response_time(self, value):
        self._response_time.insert(len(self._response_time) % 5, value)

    def ewma(self):
        ewma = 0
        divide = len(self._response_time)
        if len(self._response_time) == 0:
            return ewma
        for time in self._response_time:
            ewma += time
        return ewma / divide if divide != 0 else 0

    def is_rate_surpassed(self, request_time_sla_threshold_in_milliseconds):
        if self._is_rate_surpassed:
            self._is_rate_surpassed = self.ewma() >= 0.75 * request_time_sla_threshold_in_milliseconds
        else:
            self._is_rate_surpassed = self.ewma() >= request_time_sla_threshold_in_milliseconds

        return self._is_rate_surpassed


class Topology(object):
    def __init__(self, etag=0, nodes=None):
        self.etag = etag
        self.nodes = nodes if nodes is not None else []

    @staticmethod
    def convert_json_topology_to_entity(json_topology):
        topology = Topology()
        topology.etag = json_topology["Etag"]
        for node in json_topology["Nodes"]:
            topology.nodes.append(ServerNode(node['Url'], node['Database'], node['ClusterTag']))
        return topology


class NodeSelector(object):
    def __init__(self, topology):
        self.topology = topology
        self._current_node_index = 0
        self.lock = Lock()

    @property
    def current_node_index(self):
        return self._current_node_index

    def on_failed_request(self, node_index):
        topology_nodes_len = len(self.topology.nodes)
        if topology_nodes_len == 0:
            raise InvalidOperationException("Empty database topology, this shouldn't happen.")

        if node_index < topology_nodes_len - 1:
            next_node_index = node_index + 1
        else:
            next_node_index = 0

        self._current_node_index = next_node_index

    def on_update_topology(self, topology, force_update=False):
        if topology is None:
            return False

        old_topology = self.topology
        while True:
            if old_topology.etag >= topology.etag and not force_update:
                return False

            with self.lock:
                if not force_update:
                    self._current_node_index = 0

                if old_topology is self.topology:
                    self.topology = topology
                    return True

            old_topology = self.topology

    def get_current_node(self):
        if len(self.topology.nodes) == 0:
            raise InvalidOperationException("There are no nodes in the topology at all")

        return self.topology.nodes[self._current_node_index]

    def restore_node_index(self, node_index):
        current_node_index = self._current_node_index
        if current_node_index > node_index:
            with self.lock:
                if current_node_index == self._current_node_index:
                    self._current_node_index = node_index


class NodeStatus(object):
    def __init__(self, request_executor, node_index, node):
        self._request_executor = request_executor
        self.node_index = node_index
        self.node = node
        self._timer_period = 0
        self._timer = None

    def _next_timer_period(self):
        if not self._timer_period >= 60 * 5:
            self._timer_period += 0.1

        return 60 * 5 if self._timer_period >= 60 * 5 else self._timer_period

    def start_timer(self):
        Utils.start_a_timer(self._next_timer_period(), self._request_executor.check_node_status, (self,),
                            name="check_node_status", daemon=True)

    def __del__(self):
        if self._timer is not None:
            self._timer.cancel()



import unittest
import sys
import os

sys.path.append(os.path.abspath(__file__ + "/../../"))

from pyravendb.store.document_store import DocumentStore
from pyravendb.raven_operations.server_operations import *
from pyravendb.raven_operations.maintenance_operations import IndexDefinition, PutIndexesOperation


class User(object):
    def __init__(self, name, age):
        self.name = name
        self.age = age


class Dog(object):
    def __init__(self, name, owner):
        self.name = name
        self.owner = owner


class Patch(object):
    def __init__(self, patched):
        self.patched = patched


class TestBase(unittest.TestCase):
    @staticmethod
    def delete_all_topology_files():
        import os
        file_list = [f for f in os.listdir(".") if f.endswith("topology")]
        for f in file_list:
            os.remove(f)

    @staticmethod
    def wait_for_database_topology(store, database_name, replication_factor=1):
        topology = store.maintenance.server.send(GetDatabaseRecordOperation(database_name))
        while len(topology["Members"]) < replication_factor:
            topology = store.maintenance.server.send(GetDatabaseRecordOperation(database_name))
        return topology

    def setUp(self):
        self.default_urls = ["http://localhost.fiddler:8084"]
        self.default_database = "NorthWindTest"
        self.store = DocumentStore(urls=self.default_urls, database=self.default_database)
        self.store.initialize()
        created = False
        while not created:
            try:
                self.store.maintenance.server.send(CreateDatabaseOperation(database_name=self.default_database))
                created = True
            except Exception as e:
                if "already exists!" in str(e):
                    self.store.maintenance.server.send(
                        DeleteDatabaseOperation(database_name=self.default_database, hard_delete=True))
                    continue
                raise
        TestBase.wait_for_database_topology(self.store, self.default_database)

        self.index_map = "from doc in docs select new{Tag = doc[\"@metadata\"][\"@collection\"]}"
        self.store.maintenance.send(PutIndexesOperation(IndexDefinition("AllDocuments", maps=self.index_map)))

    def tearDown(self):
        self.store.maintenance.server.send(DeleteDatabaseOperation(database_name="NorthWindTest", hard_delete=True))
        self.store.close()



class _CommandData(object):
    def __init__(self, key, command_type, change_vector=None, metadata=None):
        self.key = key
        self._type = command_type
        self.metadata = metadata
        self.change_vector = change_vector
        self.additionalData = None
        self.__command = True

    def to_json(self):
        raise NotImplementedError

    @property
    def command(self):
        return self.__command

    @property
    def type(self):
        return self._type


class PutCommandData(_CommandData):
    def __init__(self, key, change_vector=None, document=None, metadata=None):
        super(PutCommandData, self).__init__(key, "PUT", change_vector, metadata)
        self.document = document

    def to_json(self):
        self.document["@metadata"] = self.metadata
        data = {"Type": self._type, "Id": self.key, "Document": self.document}
        if self.change_vector:
            data["ChangeVector"] = self.change_vector
        return data


class DeleteCommandData(_CommandData):
    def __init__(self, key, change_vector=None):
        super(DeleteCommandData, self).__init__(key, "DELETE", change_vector)
        self.additionalData = None

    def to_json(self):
        data = {"Id": self.key, "ChangeVector": self.change_vector, "Type": self._type}
        if self.change_vector:
            data["ChangeVector"] = self.change_vector
        return {"Id": self.key, "Type": self._type}


class PatchCommandData(_CommandData):
    def __init__(self, key, scripted_patch, change_vector=None, patch_if_missing=None, additional_data=None,
                 debug_mode=False):
        """
        @param key: key of a document to patch.
        :type str
        @param scripted_patch: (using JavaScript) that is used to patch the document.
        :type ScriptedPatchRequest
        @param change_vector: Current document change_vector, used for concurrency checks (null to skip check).
        :type str
        @param patch_if_missing: (using JavaScript) that is used to patch a default document.
        :type ScriptedPatchRequest
        @param additional_data: Additional command data. For internal use only.
        :type dict
        @param debug_mode: Indicates in the Operation should be run in debug mode.
                           for additional information in response.
        :type bool

        """

        super(PatchCommandData, self).__init__(key, "PATCH", change_vector)
        if additional_data is None:
            additional_data = {}
        self.scripted_patch = scripted_patch
        self.patch_if_missing = patch_if_missing
        self.additional_data = additional_data
        self.debug_Mode = debug_mode

    def to_json(self):
        data = {"Id": self.key, "Type": self._type,
                "Patch": self.scripted_patch.to_json(),
                "DebugMode": self.debug_Mode}

        if self.change_vector:
            data["ChangeVector"] = self.change_vector

        if self.patch_if_missing is not None:
            data["PatchIfMissing"] = self.patch_if_missing.to_json()

        return data


class PutAttachmentCommandData(_CommandData):
    def __init__(self, document_id, name, stream, content_type, change_vector):
        """
        @param document_id: The id of the document
        @param name: Name of the attachment
        @param stream: The attachment as bytes (ex.open("file_path", "rb"))
        @param content_type: The type of the attachment (ex.image/png)
        @param change_vector: The change vector of the document
        """

        if not document_id:
            raise ValueError(document_id)
        if not name:
            raise ValueError(name)

        super(PutAttachmentCommandData, self).__init__(document_id, "AttachmentPUT", change_vector)
        self.name = name
        self.stream = stream
        self.content_type = content_type

    def to_json(self):
        data = {"Id": self.key, "Name": self.name, "ContentType": self.content_type, "Type": self._type}

        if self.change_vector:
            data["ChangeVector"] = self.change_vector

        return data


class DeleteAttachmentCommandData(_CommandData):
    def __init__(self, document_id, name, change_vector):
        """
        @param document_id: The id of the document
        @param name: Name of the attachment
        @param change_vector: The change vector of the document
        """

        if not document_id:
            raise ValueError(document_id)
        if not name:
            raise ValueError(name)

        super(DeleteAttachmentCommandData, self).__init__(document_id, "AttachmentDELETE", change_vector)
        self.name = name

    def to_json(self):
        data = {"Id": self.key, "Name": self.name, "Type": self._type}

        if self.change_vector:
            data["ChangeVector"] = self.change_vector

        return data



import sys
import os

sys.path.append(os.path.abspath(__file__ + "/../"))

from pyravendb.tests.test_base import TestBase
from pyravendb.custom_exceptions import exceptions
import unittest


class Product(object):
    def __init__(self, Id=None, name=None):
        self.Id = Id
        self.name = name


class TestDelete(TestBase):
    def setUp(self):
        super(TestDelete, self).setUp()

        with self.store.open_session() as session:
            session.store(Product("products/101", "test"))
            session.store(Product("products/10", "test"))
            session.store(Product("products/106", "test"))
            session.store(Product("products/107", "test"))
            session.save_changes()

    def tearDown(self):
        super(TestDelete, self).tearDown()
        self.delete_all_topology_files()

    def test_delete_with_key_with_save_session(self):
        with self.store.open_session() as session:
            session.delete("products/101")
            session.save_changes()
            self.assertIsNone(session.load("products/101"))

    def test_delete_with_key_without_save_session(self):
        with self.store.open_session() as session:
            session.delete("products/10")
            self.assertIsNone(session.load("products/10"))

    def test_delete_after_change_fail(self):
        with self.store.open_session() as session:
            product = session.load("products/106")
            product.name = "testing"
            with self.assertRaises(exceptions.InvalidOperationException):
                session.delete("products/106")

    def test_delete_after_change_success_with_save_session(self):
        with self.store.open_session() as session:
            product = session.load("products/107")
            product.name = "testing"
            session.delete_by_entity(product)
            session.save_changes()
            self.assertIsNone(session.load("products/107"))

    def test_delete_with_entity(self):
        product = Product("Toys")
        with self.store.open_session() as session:
            session.store(product)
            session.save_changes()

        with self.store.open_session() as session:
            product = session.load(product.Id)
            session.delete(product)
            session.save_changes()

            self.assertIsNone(session.load(product.Id))


if __name__ == "__main__":
    unittest.main()



from pyravendb.tests.test_base import *
from pyravendb.commands.raven_commands import PutDocumentCommand, GetDocumentCommand


class TestPut(TestBase):
    def tearDown(self):
        super(TestPut, self).tearDown()
        TestBase.delete_all_topology_files()

    def test_put_success(self):
        request_executor = self.store.get_request_executor()
        put_command = PutDocumentCommand(key="testing/1", document={"Name": "test", "@metadata": {}})
        request_executor.execute(put_command)
        response = request_executor.execute(GetDocumentCommand("testing/1"))
        self.assertEqual(response["Results"][0]["@metadata"]["@id"], "testing/1")

    def test_put_fail(self):
        request_executor = self.store.get_request_executor()
        with self.assertRaises(ValueError):
            command = PutDocumentCommand(key="testing/2", document="document")
            request_executor.execute(command)


if __name__ == "__main__":
    unittest.main()



import threading
from functools import partial
from pyravendb.subscriptions.data import *
from pyravendb.commands.raven_commands import GetSubscriptionsCommand
from pyravendb.tests.test_base import TestBase
from pyravendb.custom_exceptions.exceptions import SubscriptionInUseException, SubscriptionClosedException
import unittest


class User:
    def __init__(self, first_name, last_name):
        self.first_name = first_name
        self.last_name = last_name

    def __str__(self):
        return self.first_name + " " + self.last_name


class TestSubscription(TestBase):
    def setUp(self):
        super(TestSubscription, self).setUp()
        self.results = []
        self.items_count = 0
        self.expected_items_count = None
        self.event = threading.Event()
        self.ack = threading.Event()

    def process_documents(self, batch):
        self.items_count += len(batch.items)
        for b in batch.items:
            self.results.append(b.result)
        if self.items_count == self.expected_items_count:
            self.event.set()

    def acknowledge(self, batch):
        if not self.ack.is_set():
            self.ack.set()

    def tearDown(self):
        super(TestSubscription, self).tearDown()
        self.delete_all_topology_files()
        self.event.clear()
        self.ack.clear()

    def test_create_and_run_subscription_with_object(self):
        self.expected_items_count = 2
        with self.store.open_session() as session:
            session.store(User("Idan", "Shalom"))
            session.store(User("Ilay", "Shalom"))
            session.store(User("Raven", "DB"))
            session.save_changes()

        creation_options = SubscriptionCreationOptions("FROM Users where last_name='Shalom'")
        self.store.subscriptions.create(creation_options)

        request_executor = self.store.get_request_executor()
        subscriptions = request_executor.execute(GetSubscriptionsCommand(0, 1))
        self.assertGreaterEqual(len(subscriptions), 1)

        connection_options = SubscriptionWorkerOptions(subscriptions[0].subscription_name)
        self.assertEqual(len(self.results), 0)
        with self.store.subscriptions.get_subscription_worker(connection_options, object_type=User) as subscription:
            subscription.run(self.process_documents)
            self.event.wait(timeout=5)

        self.assertEqual(len(self.results), 2)
        for item in self.results:
            self.assertTrue(isinstance(item, User))

    def test_subscription_continue_to_take_documents(self):
        self.expected_items_count = 3
        with self.store.open_session() as session:
            session.store(User("Idan", "Shalom"))
            session.store(User("Ilay", "Shalom"))
            session.store(User("Raven", "DB"))
            session.save_changes()

        self.store.subscriptions.create("FROM Users where last_name='Shalom'")

        request_executor = self.store.get_request_executor()
        subscriptions = request_executor.execute(GetSubscriptionsCommand(0, 1))
        self.assertGreaterEqual(len(subscriptions), 1)

        connection_options = SubscriptionWorkerOptions(subscriptions[0].subscription_name)
        self.assertEqual(len(self.results), 0)
        with self.store.subscriptions.get_subscription_worker(connection_options) as subscription:
            subscription.confirm_callback = self.acknowledge
            subscription.run(self.process_documents)
            self.ack.wait(timeout=5)

            self.assertEqual(len(self.results), 2)
            with self.store.open_session() as session:
                session.store(User("Idan", "Shalom"))
                session.save_changes()
            self.event.wait(timeout=5)
            self.assertEqual(len(self.results), 3)

    def test_drop_subscription(self):
        self.expected_items_count = 2
        with self.store.open_session() as session:
            session.store(User("Idan", "Shalom"))
            session.store(User("Ilay", "Shalom"))
            session.store(User("Raven", "DB"))
            session.save_changes()
        try:
            creation_options = SubscriptionCreationOptions("FROM Users where last_name='Shalom'")
            name = self.store.subscriptions.create(creation_options)

            worker_options = SubscriptionWorkerOptions(name, strategy=SubscriptionOpeningStrategy.wait_for_free)
            self.assertEqual(len(self.results), 0)
            subscription = self.store.subscriptions.get_subscription_worker(worker_options, object_type=User)
            subscription.run(self.process_documents)
            self.event.wait()

            subscription_throw = self.store.subscriptions.get_subscription_worker(SubscriptionWorkerOptions(name),
                                                                                  object_type=User)

            th = subscription_throw.run(self.process_documents)
            with self.assertRaises(SubscriptionInUseException):
                th.join()

            self.items_count = 0
            self.event.clear()
            self.store.subscriptions.drop_connection(name)
            with self.store.subscriptions.get_subscription_worker(SubscriptionWorkerOptions(name),
                                                                  object_type=User) as subscription:
                sub = subscription.run(self.process_documents)
                with self.store.open_session() as session:
                    session.store(User("Idan", "Shalom"))
                    session.store(User("Ilay", "Shalom"))
                    session.save_changes()

                self.event.wait()

            self.assertEqual(len(self.results), 4)
            for item in self.results:
                self.assertTrue(isinstance(item, User))

        finally:
            subscription.close()
            subscription_throw.close()

    def test_subscription_with_close_when_no_docs_left(self):
        with self.store.open_session() as session:
            for _ in range(0, 5000):
                session.store(User("Idan", "Shalom"))
                session.store(User("Ilay", "Shalom"))
            session.save_changes()

        users = []
        subscription_creation_options = SubscriptionCreationOptions("From Users")
        subscription_name = self.store.subscriptions.create(subscription_creation_options)

        worker_options = SubscriptionWorkerOptions(subscription_name,
                                                   strategy=SubscriptionOpeningStrategy.take_over,
                                                   close_when_no_docs_left=True)
        with self.store.subscriptions.get_subscription_worker(worker_options) as subscription_worker:
            try:
                subscription_worker.run(
                    partial(lambda batch, x=users: x.extend([item.raw_result for item in batch.items]))).join()
            except SubscriptionClosedException:
                # That's expected
                pass

        self.assertEqual(len(users), 10000)


if __name__ == "__main__":
    unittest.main()



class PatchRequest(object):
    def __init__(self, script, values=None):
        if values is None:
            values = {}
        self.script = script
        self.values = values

    def to_json(self):
        return {"Script": self.script, "Values": self.values}



from pyravendb.tests.test_base import *
from pyravendb.commands.raven_commands import WaitForRaftIndexCommand
from pyravendb.connection.cluster_requests_executor import ClusterRequestExecutor
import unittest


class TestServerOperations(TestBase):
    def tearDown(self):
        super(TestServerOperations, self).tearDown()
        TestBase.delete_all_topology_files()

    def test_create_database_name_longer_than_260_chars(self):
        name = "long_database_name_" + ''.join(['z' for _ in range(100)])
        self.store.maintenance.server.send(CreateDatabaseOperation(database_name=name))
        TestBase.wait_for_database_topology(self.store, name)
        database_names = self.store.maintenance.server.send(GetDatabaseNamesOperation(0, 3))
        self.assertTrue(name in database_names)

    def test_cannot_create_database_with_the_same_name(self):
        name = "Duplicate"
        self.store.maintenance.server.send(CreateDatabaseOperation(database_name=name))
        TestBase.wait_for_database_topology(self.store, name)
        with self.assertRaises(Exception):
            self.store.maintenance.server.send(CreateDatabaseOperation(database_name=name))


if __name__ == "__main__":
    unittest.main()



from pyravendb.tools.generate_id import GenerateEntityIdOnTheClient
from pyravendb.store.session_query import Query
from pyravendb.commands import commands_data
from pyravendb.commands.raven_commands import *
from pyravendb.raven_operations.operations import GetAttachmentOperation
from pyravendb.commands.commands_data import PutAttachmentCommandData, DeleteAttachmentCommandData, DeleteCommandData
from pyravendb.tools.utils import Utils
from pyravendb.data.operation import AttachmentType


class _SaveChangesData(object):
    def __init__(self, commands, deferred_command_count, entities=None):
        self.commands = commands
        self.entities = [] if entities is None else entities
        self.deferred_command_count = deferred_command_count


class DocumentSession(object):
    def __init__(self, database, document_store, requests_executor, session_id, **kwargs):
        """
        Implements Unit of Work for accessing the RavenDB server

        @param str database: the name of the database we open a session to
        @param DocumentStore document_store: the store that we work on
        """
        self.session_id = session_id
        self.database = database
        self._document_store = document_store
        self._requests_executor = requests_executor
        self._documents_by_id = {}
        self._included_documents_by_id = {}
        self._deleted_entities = set()
        self._documents_by_entity = {}
        self._known_missing_ids = set()
        self.id_value = None
        self._defer_commands = set()
        self._number_of_requests_in_session = 0
        self._query = None
        self.use_optimistic_concurrency = kwargs.get("use_optimistic_concurrency", False)
        self.advanced = Advanced(self)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    @property
    def requests_executor(self):
        return self._requests_executor

    @property
    def number_of_requests_in_session(self):
        return self._number_of_requests_in_session

    @property
    def documents_by_entity(self):
        return self._documents_by_entity

    @property
    def deleted_entities(self):
        return self._deleted_entities

    @property
    def documents_by_id(self):
        return self._documents_by_id

    def defer(self, command, *args):
        self._defer_commands.add(command)
        for arg in args:
            self._defer_commands.add(arg)

    @property
    def defer_commands(self):
        return self._defer_commands

    @property
    def known_missing_ids(self):
        return self._known_missing_ids

    @property
    def included_documents_by_id(self):
        return self._included_documents_by_id

    @property
    def conventions(self):
        return self._document_store.conventions

    @property
    def query(self):
        if self._query is None:
            self._query = Query(self)
        return self._query

    def save_includes(self, includes=None):
        if includes:
            for key, value in includes.items():
                if key not in self._documents_by_id:
                    self._included_documents_by_id[key] = value

    def save_entity(self, key, entity, original_metadata, metadata, document, force_concurrency_check=False):
        if key is not None:
            self._known_missing_ids.discard(key)

            if key not in self._documents_by_id:
                self._documents_by_id[key] = entity

                self._documents_by_entity[self._documents_by_id[key]] = {
                    "original_value": document.copy(), "metadata": metadata,
                    "original_metadata": original_metadata, "change_vector": metadata.get("change_vector", None),
                    "key": key, "force_concurrency_check": force_concurrency_check}

    def _convert_and_save_entity(self, key, document, object_type, nested_object_types):
        if key not in self._documents_by_id:
            entity, metadata, original_metadata = Utils.convert_to_entity(document, object_type, self.conventions,
                                                                          nested_object_types)
            self.save_entity(key, entity, original_metadata, metadata, document)

    def _multi_load(self, keys, object_type, includes, nested_object_types):
        if len(keys) == 0:
            return []

        ids_of_not_existing_object = set(keys)
        if not includes:
            ids_in_includes = [key for key in ids_of_not_existing_object if key in self._included_documents_by_id]
            if len(ids_in_includes) > 0:
                for include in ids_in_includes:
                    self._convert_and_save_entity(include, self._included_documents_by_id[include], object_type,
                                                  nested_object_types)
                    self._included_documents_by_id.pop(include)

            ids_of_not_existing_object = [key for key in ids_of_not_existing_object if
                                          key not in self._documents_by_id]

        ids_of_not_existing_object = [key for key in ids_of_not_existing_object if key not in self._known_missing_ids]

        if len(ids_of_not_existing_object) > 0:
            self.increment_requests_count()
            command = GetDocumentCommand(ids_of_not_existing_object, includes)
            response = self._requests_executor.execute(command)
            if response:
                results = response["Results"]
                includes = response["Includes"]
                for i in range(0, len(results)):
                    if results[i] is None:
                        self._known_missing_ids.add(ids_of_not_existing_object[i])
                        continue
                    self._convert_and_save_entity(ids_of_not_existing_object[i], results[i], object_type,
                                                  nested_object_types)
                self.save_includes(includes)
        return [None if key in self._known_missing_ids else self._documents_by_id[
            key] if key in self._documents_by_id else None for key in keys]

    def load(self, key_or_keys, object_type=None, includes=None, nested_object_types=None):
        """
        @param key_or_keys: Identifier of a document that will be loaded.
        :type str or list
        @param includes: The path to a reference inside the loaded documents can be list (property name)
        :type list or str
        @param object_type: The class we want to get
        :type classObj:
        @param nested_object_types: A dict of classes for nested object the key will be the name of the class and the
         value will be the object we want to get for that attribute
        :type str
        @return: instance of object_type or None if document with given Id does not exist.
        :rtype:object_type or None
        """
        if not key_or_keys:
            raise ValueError("None or empty key is invalid")
        if includes and not isinstance(includes, list):
            includes = [includes]

        if isinstance(key_or_keys, list):
            return self._multi_load(key_or_keys, object_type, includes, nested_object_types)

        if key_or_keys in self._known_missing_ids:
            return None
        if key_or_keys in self._documents_by_id and not includes:
            return self._documents_by_id[key_or_keys]

        if key_or_keys in self._included_documents_by_id:
            self._convert_and_save_entity(key_or_keys, self._included_documents_by_id[key_or_keys], object_type,
                                          nested_object_types)
            self._included_documents_by_id.pop(key_or_keys)
            if not includes:
                return self._documents_by_id[key_or_keys]

        self.increment_requests_count()
        command = GetDocumentCommand(key_or_keys, includes=includes)
        response = self._requests_executor.execute(command)
        if response:
            result = response["Results"]
            includes = response["Includes"]
            if len(result) == 0 or result[0] is None:
                self._known_missing_ids.add(key_or_keys)
                return None
            self._convert_and_save_entity(key_or_keys, result[0], object_type, nested_object_types)
            self.save_includes(includes)
        return self._documents_by_id[key_or_keys] if key_or_keys in self._documents_by_id else None

    def delete_by_entity(self, entity):
        if entity is None:
            raise ValueError("None entity is invalid")
        if entity not in self._documents_by_entity:
            raise exceptions.InvalidOperationException(
                "{0} is not associated with the session, cannot delete unknown entity instance".format(entity))
        if "Raven-Read-Only" in self._documents_by_entity[entity]["original_metadata"]:
            raise exceptions.InvalidOperationException(
                "{0} is marked as read only and cannot be deleted".format(entity))
        self._included_documents_by_id.pop(self._documents_by_entity[entity]["key"], None)
        self._known_missing_ids.add(self._documents_by_entity[entity]["key"])
        self._deleted_entities.add(entity)

    def delete(self, key_or_entity, expected_change_vector=None):
        """
        @param str or object key_or_entity: Can be the key or the entity we like to delete
        @param str expected_change_vector: A change vector to use for concurrency checks
        :type str or object:
        """
        if key_or_entity is None:
            raise ValueError("None key is invalid")
        if not isinstance(key_or_entity, str):
            self.delete_by_entity(key_or_entity)
            return
        if key_or_entity in self._documents_by_id:
            entity = self._documents_by_id[key_or_entity]
            if self._has_change(entity):
                raise exceptions.InvalidOperationException(
                    "Can't delete changed entity using identifier. Use delete_by_entity(entity) instead.")
            if entity not in self._documents_by_entity:
                raise exceptions.InvalidOperationException(
                    "{0} is not associated with the session, cannot delete unknown entity instance".format(entity))
            if "Raven-Read-Only" in self._documents_by_entity[entity]["original_metadata"]:
                raise exceptions.InvalidOperationException(
                    "{0} is marked as read only and cannot be deleted".format(entity))
            self.delete_by_entity(entity)
            return
        self._known_missing_ids.add(key_or_entity)
        self._included_documents_by_id.pop(key_or_entity, None)
        self.defer(commands_data.DeleteCommandData(key_or_entity, expected_change_vector))

    def assert_no_non_unique_instance(self, entity, key):
        if not (key is None or key.endswith("/") or key not in self._documents_by_id
                or self._documents_by_id[key] is entity):
            raise exceptions.NonUniqueObjectException(
                "Attempted to associate a different object with id '{0}'.".format(key))

    def store(self, entity, key="", change_vector=""):
        """
        @param entity: Entity that will be stored
        :type object:
        @param key: Entity will be stored under this key, (None to generate automatically)
        :type str:
        @param change_vector: Current entity change_vector, used for concurrency checks (null to skip check)
        :type str
        """

        if entity is None:
            raise ValueError("None entity value is invalid")

        force_concurrency_check = self._get_concurrency_check_mode(entity, key, change_vector)
        if entity in self._documents_by_entity:
            if change_vector:
                self._documents_by_entity[entity]["change_vector"] = change_vector
            self._documents_by_entity[entity]["force_concurrency_check"] = force_concurrency_check
            return

        if not key:
            entity_id = GenerateEntityIdOnTheClient.try_get_id_from_instance(entity)
        else:
            GenerateEntityIdOnTheClient.try_set_id_on_entity(entity, key)
            entity_id = key

        self.assert_no_non_unique_instance(entity, entity_id)

        if not entity_id:
            entity_id = self._document_store.generate_id(self.database, entity)
            GenerateEntityIdOnTheClient.try_set_id_on_entity(entity, entity_id)

        for command in self._defer_commands:
            if command.key == entity_id:
                raise exceptions.InvalidOperationException(
                    "Can't store document, there is a deferred command registered for this document in the session. "
                    "Document id: " + entity_id)

        if entity in self._deleted_entities:
            raise exceptions.InvalidOperationException(
                "Can't store object, it was already deleted in this session.  Document id: " + entity_id)

        metadata = self.conventions.build_default_metadata(entity)
        self._deleted_entities.discard(entity)
        self.save_entity(entity_id, entity, {}, metadata, {}, force_concurrency_check=force_concurrency_check)

    def _get_concurrency_check_mode(self, entity, key="", change_vector=""):
        if key == "":
            return "disabled" if change_vector is None else "forced"
        if change_vector == "":
            if key is None:
                entity_key = GenerateEntityIdOnTheClient.try_get_id_from_instance(entity)
                return "forced" if entity_key is None else "auto"
            return "auto"

        return "disabled" if change_vector is None else "forced"

    def save_changes(self):
        data = _SaveChangesData(list(self._defer_commands), len(self._defer_commands))
        self._defer_commands.clear()
        self._prepare_for_delete_commands(data)
        self._prepare_for_puts_commands(data)
        if len(data.commands) == 0:
            return
        self.increment_requests_count()
        batch_command = BatchCommand(data.commands)
        batch_result = self._requests_executor.execute(batch_command)
        if batch_result is None:
            raise exceptions.InvalidOperationException(
                "Cannot call Save Changes after the document store was disposed.")
        self._update_batch_result(batch_result, data)

    def _update_batch_result(self, batch_result, data):

        i = data.deferred_command_count
        batch_result_length = len(batch_result)
        while i < batch_result_length:
            item = batch_result[i]
            if item["Type"] == "PUT":
                entity = data.entities[i - data.deferred_command_count]
                if entity in self._documents_by_entity:
                    self._documents_by_id[item["@id"]] = entity
                    document_info = self._documents_by_entity[entity]
                    document_info["change_vector"] = item["@change-vector"]
                    item.pop("Type", None)
                    document_info["original_metadata"] = item.copy()
                    document_info["metadata"] = item
                    document_info["original_value"] = entity.__dict__.copy()
            i += 1

    def _prepare_for_delete_commands(self, data):
        keys_to_delete = []
        for entity in self._deleted_entities:
            keys_to_delete.append(self._documents_by_entity[entity]["key"])

        for key in keys_to_delete:
            change_vector = None
            if key in self._documents_by_id:
                existing_entity = self._documents_by_id[key]
                if existing_entity in self._documents_by_entity:
                    change_vector = self._documents_by_entity[existing_entity]["metadata"][
                        "@change-vector"] if self.advanced.use_optimistic_concurrency else None
                self._documents_by_entity.pop(existing_entity, None)
                self._documents_by_id.pop(key, None)
                data.entities.append(existing_entity)
            data.commands.append(commands_data.DeleteCommandData(key, change_vector))
        self._deleted_entities.clear()

    def _prepare_for_puts_commands(self, data):
        for entity in self._documents_by_entity:
            if self._has_change(entity):
                key = self._documents_by_entity[entity]["key"]
                metadata = self._documents_by_entity[entity]["metadata"]
                change_vector = None
                if self.advanced.use_optimistic_concurrency and self._documents_by_entity[entity][
                    "force_concurrency_check"] != "disabled" or self._documents_by_entity[entity][
                    "force_concurrency_check"] == "forced":
                    change_vector = self._documents_by_entity[entity]["change_vector"]
                data.entities.append(entity)
                if key:
                    self._documents_by_id.pop(key)
                    document = entity.__dict__.copy()
                    document.pop('Id', None)
                data.commands.append(commands_data.PutCommandData(key, change_vector, document, metadata))

    def _has_change(self, entity):
        if self._documents_by_entity[entity]["original_metadata"] != self._documents_by_entity[entity]["metadata"] \
                or self._documents_by_entity[entity]["original_value"] != entity.__dict__:
            return True
        return False

    def increment_requests_count(self):
        self._number_of_requests_in_session += 1
        if self._number_of_requests_in_session > self.conventions.max_number_of_request_per_session:
            raise exceptions.InvalidOperationException(
                "The maximum number of requests ({0}) allowed for this session has been reached. Raven limits the number \
                of remote calls that a session is allowed to make as an early warning system. Sessions are expected to \
                be short lived, and Raven provides facilities like batch saves (call save_changes() only once).\
                You can increase the limit by setting DocumentConventions.\
                MaxNumberOfRequestsPerSession or MaxNumberOfRequestsPerSession, but it is advisable \
                that you'll look into reducing the number of remote calls first, \
                since that will speed up your application significantly and result in a\
                more responsive application.".format(self.conventions.max_number_of_request_per_session))


class Advanced(object):

    def __init__(self, session):
        self.session = session
        self.use_optimistic_concurrency = session.use_optimistic_concurrency if session.use_optimistic_concurrency \
            else session.conventions.use_optimistic_concurrency
        self._attachment = None

    @property
    def attachment(self):
        if not self._attachment:
            self._attachment = _Attachment(self.session)
        return self._attachment

    def stream(self, query):
        from pyravendb.store.stream import IncrementalJsonParser
        import ijson

        index_query = query.get_index_query()
        if index_query.wait_for_non_stale_results:
            raise exceptions.NotSupportedException("Since stream() does not wait for indexing (by design), "
                                                   "streaming query with wait_for_non_stale_results is not supported.")
        self.session.increment_requests_count()
        command = QueryStreamCommand(index_query)
        with self.session.requests_executor.execute(command) as response:
            basic_parse = IncrementalJsonParser.basic_parse(response)
            parser = ijson.backend.common.parse(basic_parse)

            results = ijson.backend.common.items(parser, "Results")
            for result in next(results, None):
                document, metadata, _ = Utils.convert_to_entity(result, query.object_type, self.session.conventions,
                                                                query.nested_object_types)
                yield {"document": document, "metadata": metadata, "id": metadata.get("@id", None),
                       "change-vector": metadata.get("@change-vector", None)}

    def number_of_requests_in_session(self):
        return self.session.number_of_requests_in_session

    def get_document_id(self, instance):
        if instance is not None:
            if instance in self.session.documents_by_entity:
                return self.session.documents_by_entity[instance]["key"]
        return None

    @property
    def document_store(self):
        """
        The document store associated with this session
        """
        return self.session._document_store

    def clear(self):
        self.session.documents_by_entity.clear()
        self.session.documents_by_id.clear()
        self.session.deleted_entities.clear()
        self.session.included_documents_by_id.clear()
        self.session.known_missing_ids.clear()


class _Attachment:
    def __init__(self, session):
        self._session = session
        self._store = session.advanced.document_store
        self._defer_commands = session.defer_commands

    def _command_exists(self, document_id):
        for command in self._defer_commands:
            if not isinstance(command, (PutAttachmentCommandData, DeleteAttachmentCommandData, DeleteCommandData)):
                continue

            if command.key == document_id:
                return command

    def throw_not_in_session(self, entity):
        raise ValueError(
            repr(entity) + " is not associated with the session, cannot add attachment to it. "
                           "Use documentId instead or track the entity in the session.")

    def store(self, entity_or_document_id, name, stream, content_type=None, change_vector=None):
        if not isinstance(entity_or_document_id, str):
            entity = self._session.documents_by_entity.get(entity_or_document_id, None)
            if not entity:
                self.throw_not_in_session(entity_or_document_id)
            entity_or_document_id = entity["metadata"]["@id"]

        if not entity_or_document_id:
            raise ValueError(entity_or_document_id)
        if not name:
            raise ValueError(name)

        defer_command = self._command_exists(entity_or_document_id)
        if defer_command:
            message = "Can't store attachment {0} of document {1}".format(name, entity_or_document_id)
            if defer_command.type == "DELETE":
                message += ", there is a deferred command registered for this document to be deleted."
            elif defer_command.type == "AttachmentPUT":
                message += ", there is a deferred command registered to create an attachment with the same name."
            elif defer_command.type == "AttachmentDELETE":
                message += ", there is a deferred command registered to delete an attachment with the same name."
            raise exceptions.InvalidOperationException(message)

        entity = self._session.documents_by_id.get(entity_or_document_id, None)
        if entity and entity in self._session.deleted_entities:
            raise exceptions.InvalidOperationException(
                "Can't store attachment " + name + " of document" + entity_or_document_id +
                ", the document was already deleted in this session.")

        self._session.defer(PutAttachmentCommandData(entity_or_document_id, name, stream, content_type, change_vector))

    def delete(self, entity_or_document_id, name):
        if not isinstance(entity_or_document_id, str):
            entity = self._session.documents_by_entity.get(entity_or_document_id, None)
            if not entity:
                self.throw_not_in_session(entity_or_document_id)
            entity_or_document_id = entity["metadata"]["@id"]

        if not entity_or_document_id:
            raise ValueError(entity_or_document_id)
        if not name:
            raise ValueError(name)

        defer_command = self._command_exists(entity_or_document_id)
        if defer_command:
            if defer_command.type in ("Delete", "AttachmentDELETE"):
                return
            if defer_command.type == "AttachmentPUT":
                raise exceptions.InvalidOperationException(
                    "Can't delete attachment " + name + " of document " + entity_or_document_id +
                    ",there is a deferred command registered to create an attachment with the same name.")

        entity = self._session.documents_by_id.get(entity_or_document_id, None)
        if entity and entity in self._session.deleted_entities:
            return

        self._session.defer(DeleteAttachmentCommandData(entity_or_document_id, name, None))

    def get(self, entity_or_document_id, name):
        if not isinstance(entity_or_document_id, str):
            entity = self._session.documents_by_entity.get(entity_or_document_id, None)
            if not entity:
                self.throw_not_in_session(entity_or_document_id)
            entity_or_document_id = entity["metadata"]["@id"]

        operation = GetAttachmentOperation(entity_or_document_id, name, AttachmentType.document, None)
        return self._store.operations.send(operation)



from pyravendb.data.indexes import SortOptions
from datetime import datetime, timedelta
from pyravendb.tools.utils import Utils
from inflector import Inflector

inflector = Inflector()


class DocumentConventions(object):
    def __init__(self, **kwargs):
        self.max_number_of_request_per_session = kwargs.get("max_number_of_request_per_session", 30)
        self.max_ids_to_catch = kwargs.get("max_ids_to_catch", 32)
        # timeout for wait to server in seconds
        self.timeout = kwargs.get("timeout", timedelta(seconds=30))
        self.use_optimistic_concurrency = kwargs.get("use_optimistic_concurrency", False)
        self.json_default_method = DocumentConventions.json_default
        self.max_length_of_query_using_get_url = kwargs.get("max_length_of_query_using_get_url", 1024 + 512)
        self.identity_parts_separator = "/";
        self.disable_topology_update = kwargs.get("disable_topology_update", False)
        # If set to 'true' then it will throw an exception when any query is performed (in session)
        # without explicit page size set
        self.raise_if_query_page_size_is_not_set = kwargs.get("raise_if_query_page_size_is_not_set", False)

    @staticmethod
    def json_default(o):
        if o is None:
            return None

        if isinstance(o, datetime):
            return Utils.datetime_to_string(o)
        elif isinstance(o, timedelta):
            return Utils.timedelta_to_str(o)
        elif getattr(o, "__dict__", None):
            return o.__dict__
        elif isinstance(o, int) or isinstance(o, float):
            return str(o)
        else:
            raise TypeError(repr(o) + " is not JSON serializable (Try add a json default method to store convention)")

    @staticmethod
    def default_transform_plural(name):
        return inflector.conditional_plural(2, name)

    @staticmethod
    def default_transform_type_tag_name(name):
        count = sum(1 for c in name if c.isupper())
        if count <= 1:
            return DocumentConventions.default_transform_plural(name.lower())
        return DocumentConventions.default_transform_plural(name)

    @staticmethod
    def build_default_metadata(entity):
        if entity is None:
            return {}
        return {"@collection": DocumentConventions.default_transform_plural(entity.__class__.__name__),
                "Raven-Python-Type": "{0}.{1}".format(entity.__class__.__module__, entity.__class__.__name__)}

    @staticmethod
    def try_get_type_from_metadata(metadata):
        if "Raven-Python-Type" in metadata:
            return metadata["Raven-Python-Type"]
        return None

    @staticmethod
    def uses_range_type(obj):
        if obj is None:
            return False

        if isinstance(obj, int) or isinstance(obj, float):
            return True
        return False

    @staticmethod
    def get_default_sort_option(type_name):
        if not type_name:
            return None
        if type_name == "int" or type_name == "float" or type_name == "long":
            return SortOptions.numeric

    @staticmethod
    def range_field_name(field_name, type_name):
        if type_name == "long":
            field_name = "{0}_L_Range".format(field_name)
        else:
            field_name = "{0}_D_Range".format(field_name)

        return field_name



import ijson
import re

LEXEME_RE = re.compile(r'[a-z0-9eE\.\+-]+|\S')


# The code imported from ijson to be able to receive json from socket
class IncrementalJsonParser:
    def __init__(self, socket):
        self.lexer = self.lexer(socket)

    @staticmethod
    def basic_parse(response, buf_size=ijson.backend.BUFSIZE):
        """
        Iterator yielding unprefixed events.

        Parameters:

        - response: a stream response from requests
        """
        lexer = iter(IncrementalJsonParser.lexer(response, buf_size))
        for value in ijson.backend.parse_value(lexer):
            yield value
        try:
            next(lexer)
        except StopIteration:
            pass
        else:
            raise ijson.common.JSONError('Additional data')

    @staticmethod
    def lexer(response, buf_size=ijson.backend.BUFSIZE):
        it = response.iter_content(buf_size)
        data = next(it, None)
        if data is None:
            yield None
        buf = data.decode('utf-8')
        pos = 0
        discarded = 0
        while True:
            match = LEXEME_RE.search(buf, pos)
            if pos < len(buf) and match:
                lexeme = match.group()
                if lexeme == '"':
                    pos = match.start()
                    start = pos + 1
                    while True:
                        try:
                            end = buf.index('"', start)
                            escpos = end - 1
                            while buf[escpos] == '\\':
                                escpos -= 1
                            if (end - escpos) % 2 == 0:
                                start = end + 1
                            else:
                                break
                        except ValueError:
                            data = next(it, buf_size)
                            if not data:
                                yield None
                            buf += data.decode('utf-8')
                    yield discarded + pos, buf[pos:end + 1]
                    pos = end + 1
                else:
                    while match.end() == len(buf):
                        data = next(it, None)
                        if not data:
                            break
                        buf += data.decode('utf-8')
                        match = LEXEME_RE.search(buf, pos)
                        lexeme = match.group()
                    pos = match.end()
                    yield discarded + match.start(), lexeme
            else:
                data = next(it, None)
                if data is None:
                    yield None
                buf = data.decode('utf-8')
                pos = 0



from datetime import timedelta
from enum import Enum
import sys


class IndexLockMode(Enum):
    unlock = "Unlock"
    locked_ignore = "LockedIgnore"
    locked_error = "LockedError"
    side_by_side = "SideBySide"

    def __str__(self):
        return self.value


class IndexPriority(Enum):
    low = "Low"
    normal = "Normal"
    high = "High"

    def __str__(self):
        return self.value


# The sort options to use for a particular field
class SortOptions(Enum):
    # No sort options
    none = "None"
    # Sort using term values as Strings. Sort values are str and lower values are at the front.
    str = "String"
    # Sort using term values as encoded Doubles and Longs.
    # Sort values are float or longs and lower values are at the front.
    numeric = "Numeric"

    def __str__(self):
        return self.value


class FieldIndexing(Enum):
    # Do not index the field value.
    no = "No"
    # Index the tokens produced by running the field's value through an Analyzer
    search = "Search"
    # Index the field's value without using an Analyzer, so it can be searched.
    exact = "Exact"
    # Index this field using the default internal analyzer: LowerCaseKeywordAnalyzer
    default = "Default"

    def __str__(self):
        return self.value


class FieldTermVector(Enum):
    # Do not store term vectors
    no = "No"
    # Store the term vectors of each document.
    # A term vector is a list of the document's terms and their number of occurrences in that document.
    yes = "Yes"
    # store the term vector + token position information
    with_positions = "WithPositions"
    # Store the term vector + Token offset information
    with_offsets = "WithOffsets"
    # Store the term vector + Token position and offset information
    with_positions_and_offsets = "WithPositionsAndOffsets"

    def __str__(self):
        return self.value


class IndexDefinition(object):
    def __init__(self, name, maps, configuration=None, **kwargs):
        """
        @param name: The name of the index
        :type str
        @param maps: The map of the index
        :type str ,set or list of str
        @param kwargs: Can be use to initialize the other option in the index definition
        :type kwargs
        """
        self.name = name
        self.configuration = configuration if configuration is not None else {}
        self.reduce = kwargs.get("reduce", None)

        self.index_id = kwargs.get("index_id", 0)
        self.is_test_index = kwargs.get("is_test_index", False)

        # IndexLockMode
        self.lock_mod = kwargs.get("lock_mod", None)

        # The priority of the index IndexPriority
        self.priority = kwargs.get("priority", None)
        if isinstance(maps, str):
            self.maps = (maps,)
        else:
            if isinstance(maps, list):
                maps = set(maps)
            self.maps = tuple(maps, )

        # fields is a  key value dict. the key is the name of the field and the value is IndexFieldOptions
        self.fields = kwargs.get("fields", {})

    @property
    def type(self):
        value = "Map"
        if self.name and self.name.startswith('Auto/'):
            value = "AutoMap"
            if self.reduce:
                value += "Reduce"
        elif self.reduce:
            return "MapReduce"
        return value

    @property
    def is_map_reduce(self):
        return True if self.reduce else False

    @property
    def map(self):
        if not isinstance(self.maps, str):
            return self.maps[0]
        return self.maps

    @map.setter
    def map(self, value):
        if len(self.maps) != 0:
            self.maps.pop()
        self.maps.add(value)

    def to_json(self):
        return {"Configuration": self.configuration,
                "Fields": {key: self.fields[key].to_json() for key in self.fields} if len(
                    self.fields) > 0 else self.fields,
                "IndexId": self.index_id,
                "IsTestIndex": self.is_test_index,
                "LockMode": str(self.lock_mod) if self.lock_mod else None,
                "Maps": self.maps,
                "Name": self.name,
                "Reduce": self.reduce,
                "OutputReduceToCollection": None,
                "Priority": str(self.priority) if self.priority else None,
                "Type": self.type}


class IndexFieldOptions(object):
    def __init__(self, sort_options=None, indexing=None, storage=None, suggestions=None, term_vector=None,
                 analyzer=None):
        """
        @param SortOptions sort_options: Sort options to use for a particular field
        @param FieldIndexing indexing: Options for indexing a field
        @param bool storage: Specifies whether a field should be stored
        @param bool suggestions: If to produce a suggestions in query
        @param FieldTermVector term_vector: Specifies whether to include term vectors for a field
        @param str analyzer: To make an entity property indexed using a specific Analyzer (Need the name)
        """
        self.sort_options = sort_options
        self.indexing = indexing
        self.storage = storage
        self.suggestions = suggestions
        self.term_vector = term_vector
        self.analyzer = analyzer

    def to_json(self):
        return {"Analyzer": self.analyzer,
                "Indexing": str(self.indexing) if self.indexing else None,
                "Sort": str(self.sort_options) if self.sort_options else None,
                "Spatial": None,
                "Storage": None if self.storage is None else "Yes" if self.storage == True else "No",
                "Suggestions": self.suggestions,
                "TermVector": str(self.term_vector) if self.term_vector else None}



from pyravendb.tests.test_base import *
from pyravendb.raven_operations.maintenance_operations import *
import unittest


class TestMaintenanceOperations(TestBase):
    def tearDown(self):
        super(TestMaintenanceOperations, self).tearDown()
        TestBase.delete_all_topology_files()

    def test_put_index_operation(self):
        index_definition_users = IndexDefinition("Users", "from doc in docs.Users select new {doc.Name}")
        index_definition_dogs = IndexDefinition("Dogs", "from doc in docs.Dogs select new {doc.brand}")

        self.store.maintenance.send(PutIndexesOperation(index_definition_dogs, index_definition_users))
        index_names = self.store.maintenance.send(GetIndexNamesOperation(0, 25))
        self.assertTrue("Users" in index_names and "Dogs" in index_names)

    def test_get_index_operation(self):
        index_definition_users = IndexDefinition("Users", "from doc in docs.Users select new {doc.Name}")
        self.store.maintenance.send(PutIndexesOperation(index_definition_users))
        users_index_definition = self.store.maintenance.send(GetIndexOperation("Users"))
        self.assertEqual(users_index_definition.name, "Users")


if __name__ == "__main__":
    unittest.main()



from pyravendb.tests.test_base import TestBase
from pyravendb.store.document_store import DocumentStore
from pyravendb.custom_exceptions import exceptions
import unittest


class Product(object):
    def __init__(self, Id=None, name=""):
        self.Id = Id
        self.name = name


class Company(object):
    def __init__(self, Id=None, name="", product=None):
        self.Id = Id
        self.name = name
        self.product = product


class Foo(object):
    def __init__(self, name, key):
        self.name = name
        self.key = key


class Time(object):
    def __init__(self, td):
        self.td = td


class Order(object):
    def __init__(self, Id=None, name="", key=None, product_id=None):
        self.Id = Id
        self.name = name
        self.key = key
        self.product_id = product_id


class TestLoad(TestBase):
    def setUp(self):
        super(TestLoad, self).setUp()

        with self.store.open_session() as session:
            session.store(Product("products/101", "test"))
            session.store(Product("products/10", "test"))
            session.store(Order("orders/105", "testing_order", 92, "products/101"))
            session.store(Company("company/1", "test", Product(name="testing_nested")))
            session.save_changes()

    def tearDown(self):
        super(TestLoad, self).tearDown()
        self.delete_all_topology_files()

    def test_load_success(self):
        with self.store.open_session() as session:
            product = session.load("products/101")
            self.assertEqual(product.name, "test")

    def test_load_missing_document(self):
        with self.store.open_session() as session:
            self.assertEqual(session.load("products/0"), None)

    def test_multi_load(self):
        with self.store.open_session() as session:
            products = session.load(["products/101", "products/10"])
            self.assertEqual(len(products), 2)

    def test_multi_load_with_duplicate_id(self):
        with self.store.open_session() as session:
            products = session.load(["products/101", "products/101", "products/10"])
            self.assertEqual(len(products), 3)
            for product in products:
                self.assertIsNotNone(product)

    def test_load_track_entity(self):
        with self.store.open_session() as session:
            product = session.load("products/101")
            self.assertTrue(isinstance(product, Product))

    def test_load_track_entity_with_object_type(self):
        with self.store.open_session() as session:
            product = session.load("products/101", object_type=Product)
            self.assertTrue(isinstance(product, Product))

    def test_load_track_entity_with_object_type_and_nested_object(self):
        with self.store.open_session() as session:
            company = session.load("company/1", object_type=Company, nested_object_types={"product": Product})
            self.assertTrue(isinstance(company, Company) and isinstance(company.product, Product))

    def test_load_track_entity_with_object_type_fail(self):
        with self.store.open_session() as session:
            with self.assertRaises(exceptions.InvalidOperationException):
                session.load("products/101", object_type=Foo)

    def test_load_with_include(self):
        with self.store.open_session() as session:
            session.load("orders/105", includes="product_id")
            session.load("products/101")
        self.assertEqual(session.number_of_requests_in_session, 1)


if __name__ == "__main__":
    unittest.main()



import logging
import socket
import ssl
import sys
from datetime import datetime
from pyravendb.tools.utils import Utils
from pyravendb.connection.requests_helpers import PropagatingThread
from urllib.parse import urlparse
from pyravendb.commands.raven_commands import GetTcpInfoCommand
from pyravendb.subscriptions.data import IncrementalJsonParser
from pyravendb.custom_exceptions.exceptions import *
from pyravendb.connection.requests_executor import RequestsExecutor
from pyravendb.custom_exceptions.exceptions import AggregateException
import ijson


class SubscriptionWorker:
    def __init__(self, options, store, database_name, confirm_callback=None, on_subscription_connection_retry=None,
                 object_type=None, nested_object_types=None):
        """
        @param SubscriptionWorkerOptions options: The subscription connection options.
        @param DocumentStore store: The document store
        @param str database_name: The database name
        @param func(SubscriptionBatch) confirm_callback: Allows the user to define stuff that happens after the confirm was received from the server
        @param func(Exception) on_subscription_connection_retry: Allows the user to define what will happen after connection retry
        @param Type object_type: The type of the object we want to track the entity to
        @param Dict[str, Type] nested_object_types: A dict of classes for nested object the key will be the name of the
        class and the value will be the object we want to get for that attribute
        """
        self._options = options
        self._store = store
        self._database_name = database_name if database_name is not None else store.database
        self._logger = logging.getLogger("subscription_" + self._database_name)
        handler = logging.FileHandler('subscriptions.log')
        formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
        handler.setFormatter(formatter)
        self._logger.addHandler(handler)
        self._logger.setLevel(logging.DEBUG)

        if options.subscription_name is None:
            raise AttributeError("SubscriptionWorkerOptions must specify the subscription_name", "options")

        self._deleted = False
        self._process_documents = None
        self._my_socket = None
        self.request_executor = None
        self._subscription_thread = None
        self._closed = False
        self.object_type = object_type
        self.nested_object_types = nested_object_types
        self._batch = None
        self.confirm_callback = confirm_callback
        self._redirect_node = None

        self.on_subscription_connection_retry = on_subscription_connection_retry
        self._last_connection_failure = None

    def connect_to_server(self):
        command = GetTcpInfoCommand("Subscription/" + self._database_name, self._database_name)
        request_executor = self._store.get_request_executor(self._database_name)

        send_buffer_size = 32 * 1024
        receive_buffer_size = 4096

        if self._redirect_node:
            try:
                result = request_executor.execute_with_node(self._redirect_node, command, should_retry=False)
            except Exception:
                # if we failed to talk to a node, we'll forget about it and let the topology to
                # redirect us to the current nod
                self._redirect_node = None
                raise
        else:
            result = request_executor.execute(command)

        uri = urlparse(result["Url"])
        host = uri.hostname
        port = uri.port

        self._my_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._my_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        self._my_socket.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, send_buffer_size)
        self._my_socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, receive_buffer_size)
        self._my_socket.settimeout(30)

        # TODO: wrap SSL
        if self._store.certificate:
            self._my_socket = ssl.wrap_socket(self._my_socket, certfile=self._store.certificate,
                                              ssl_version=ssl.PROTOCOL_TLSv1_2)
        try:
            self._my_socket.connect((host, port))
        except Exception as e:
            print(e)

        header = Utils.dict_to_bytes(
            {"Operation": "Subscription", "DatabaseName": self._database_name, "OperationVersion": 40})
        self._my_socket.sendall(header)

        parser = IncrementalJsonParser(self._my_socket)
        response = parser.next_object()

        if response['Status'] != "Ok":
            if response['Status'] == "AuthorizationFailed":
                raise ConnectionRefusedError(
                    "Cannot access database {0} because {1}".format(self._database_name, response['Message']))
            elif response['Status'] == "TcpVersionMismatch":
                raise InvalidOperationException(
                    "Cannot access database {0} because {1}".format(self._database_name, response['Message']))

        options = Utils.dict_to_bytes(self._options.to_json())
        self._my_socket.sendall(options)

        if self.request_executor:
            self.request_executor.close()
        self.request_executor = RequestsExecutor.create_for_single_node(command.requested_node.url,
                                                                        self._database_name,
                                                                        self._store.certificate)

        return parser

    def read_single_batch_from_server(self, parser):
        while True and not self._closed:
            item = parser.next_object()
            if item is None:
                break
            if item['Type'] == 'Data':
                self._batch.raw_items.append(item)
                continue
            if item['Type'] == 'EndOfBatch':
                return
            if item['Type'] == 'Confirm':
                if self.confirm_callback:
                    self.confirm_callback(self._batch)
                self._batch.clear()
                continue

            if item['Type'] == 'ConnectionStatus':
                self.assert_connection_state(item)
            if item['Type'] == 'Error':
                err = item.get('Exception', 'None')
                raise InvalidOperationException("Connection terminated by server Exception: " + err)

            raise InvalidOperationException("Unrecognized message from server: " + item['Type'])

    def run_subscription(self):
        import time
        while not self._closed:
            try:
                self.process_subscription()
            except Exception as ex:
                try:
                    if sys.gettrace():
                        self._logger.info("Subscription '{0}'. Pulling task threw the following exception : {1}".format(
                            self._options.subscription_name, ex))
                    if self.should_try_to_reconnect(ex):
                        time.sleep(self._options.time_to_wait_before_connection_retry.seconds)
                        if self.on_subscription_connection_retry:
                            self.on_subscription_connection_retry(ex)
                    else:
                        if sys.gettrace():
                            self._logger.info(
                                "Connection to subscription '{0}'. have been shut down because of an error : {1}".format(
                                    self._options.subscription_name, ex))
                        raise
                except Exception as e:
                    if e == ex:
                        raise
                    raise AggregateException(e, ex)

    def process_subscription(self):
        try:
            parser = self.connect_to_server()
            if self._closed:
                return
            connection_status = parser.next_object()
            if connection_status['Type'] != "ConnectionStatus" or connection_status['Status'] != "Accepted":
                self.assert_connection_state(connection_status)

            self._batch = SubscriptionBatch(self.request_executor, self._store, self._database_name, self._logger,
                                            self.object_type, self.nested_object_types)
            while not self._closed:
                self.read_single_batch_from_server(parser)
                last_change_vector = self._batch.initialize()
                try:
                    self._process_documents(self._batch)
                except Exception as ex:
                    if sys.gettrace():
                        self._logger.info(
                            "Subscription '{0}'. Subscriber threw an exception on document batch : {1}".format(
                                self._options.subscription_name, str(ex)))
                    if not self._options.ignore_subscriber_errors:
                        raise SubscriberErrorException(
                            "Subscriber threw an exception in subscription" + self._options.subscription_name, ex)

                header = Utils.dict_to_bytes(
                    {"ChangeVector": last_change_vector, "Type": "Acknowledge"})
                self._my_socket.sendall(header)
        except (InvalidOperationException, OSError) as e:
            "This is raise when shutting down, it isn't an error, so we don't need to treat it as such"
        finally:
            if self._my_socket:
                self._my_socket.close()

    def should_try_to_reconnect(self, exception):
        if isinstance(exception, SubscriptionDoesNotBelongToNodeException):
            self.assert_last_connection_failure()
            request_executor = self._store.get_request_executor(self._database_name)
            if not exception.appropriate_node:
                return True

            node_to_redirect_to = Utils.first_or_default(request_executor.topology_nodes,
                                                         lambda node: node.cluster_tag == exception.appropriate_node,
                                                         None)

            if not node_to_redirect_to:
                raise AggregateException(exception, InvalidOperationException(
                    "Could not redirect to {0}, "
                    "because it was not found in local topology, even after retrying".format(
                        exception.appropriate_node)))

            self._redirect_node = node_to_redirect_to
            return True
        if isinstance(exception, SubscriptionChangeVectorUpdateConcurrencyException):
            return True
        if isinstance(exception, (
                SubscriptionInUseException, SubscriptionDoesNotExistException, SubscriptionClosedException,
                SubscriptionInvalidStateException, DatabaseDoesNotExistException, AuthorizationException,
                AllTopologyNodesDownException, SubscriberErrorException, ValueError, NotImplementedError,
                AttributeError, ijson.backend.UnexpectedSymbol)):
            self.close()
            return False
        self.assert_last_connection_failure()
        return True

    def assert_last_connection_failure(self):
        if not self._last_connection_failure:
            self._last_connection_failure = datetime.now()
        elif (datetime.now() - self._last_connection_failure) > self._options.max_erroneous_period:
            raise SubscriptionInvalidStateException(
                "Subscription connection was in invalid state for more than {0} and therefore will be terminated".format(
                    self._options.max_erroneous_period))

    def assert_connection_state(self, connection_status):
        if connection_status['Type'] == "Error":
            if "DatabaseDoesNotExistException" in connection_status['Exception']:
                raise "Database '{0}' does not exist.".format(self._database_name)
        if connection_status['Type'] != "ConnectionStatus":
            raise Exception(
                "Server returned illegal type message when expecting connection status, was: " + connection_status[
                    'Type'])

        status = connection_status['Status']
        if status != "Accepted":
            if status == "InUse":
                raise SubscriptionInUseException(
                    "Subscription With Id '{0}' cannot be opened, "
                    "because it's in use and the connection strategy is {1}".format(
                        self._options.subscription_name, self._options.strategy))
            if status == "Closed":
                raise SubscriptionClosedException(
                    "Subscription With Id '{0}' cannot be opened, because it is in invalid state. {1}".format(
                        self._options.subscription_name, connection_status["Exception"]))
            if status == "Invalid":
                raise SubscriptionInvalidStateException(
                    "Subscription With Id '{0}' cannot be opened, because it is in invalid state. {1}".format(
                        self._options.subscription_name, connection_status["Exception"]))
            if status == "NotFound":
                raise SubscriptionDoesNotExistException(
                    "Subscription With Id '{0}' cannot be opened, because it does not exist. {1}".format(
                        self._options.subscription_name, connection_status["Exception"]))
            if status == "Redirect":
                if "Data" in connection_status:
                    if "RedirectedTag" in connection_status["Data"]:
                        appropriate_node = connection_status["Data"]["RedirectedTag"]
                        raise SubscriptionDoesNotBelongToNodeException(
                            appropriate_node,
                            "Subscription With Id '{0}' cannot be processed by current node, it will be redirected to {1}".format(
                                self._options.subscription_name,
                                appropriate_node if appropriate_node is not None else "None"))
            if status == "ConcurrencyReconnect":
                raise SubscriptionChangeVectorUpdateConcurrencyException(connection_status["Exception"])
            raise AttributeError(
                "Subscription '{0}' could not be opened, reason: {1}".format(
                    self._options.subscription_name, status))

    def run(self, process_documents):
        if self._subscription_thread:
            raise SubscriptionInUseException("Subscription already in use")
        self._process_documents = process_documents
        self._subscription_thread = PropagatingThread(target=self.run_subscription)
        self._subscription_thread.start()
        return self._subscription_thread

    def close(self):
        try:
            if self._closed or not self._subscription_thread:
                return

            self._closed = True
            if self._my_socket:
                self._my_socket.close()
            self._subscription_thread.join()
        except Exception as ex:
            self._logger.debug("Error during closing the subscription: " + str(ex))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class _SubscriptionBatchItem:
    def __init__(self, change_vector, document_id, raw_result, raw_metadata, result, exception_message):
        self.change_vector = change_vector
        self.document_id = document_id
        self.raw_result = raw_result
        self.raw_metadata = raw_metadata
        self.result = result
        self.exception_message = exception_message


class SubscriptionBatch:
    def __init__(self, request_executor, store, database_name, logger, object_type, nested_object_type):
        self._request_executor = request_executor
        self._store = store
        self._database_name = database_name
        self._logger = logger
        self._object_type = object_type
        self._nested_object_type = nested_object_type
        self.raw_items = []
        self.items = []  # List[_SubscriptionBatchItem]

    def open_session(self):
        return self._store.open_session(self._database_name, self._request_executor)

    def clear(self):
        self.raw_items.clear()
        self.items.clear()

    def initialize(self):
        last_change_vector = None
        for item in self.raw_items:

            entity, metadata, _ = Utils.convert_to_entity(item['Data'], self._object_type,
                                                          self._store.conventions,
                                                          nested_object_types=self._nested_object_type)
            if not metadata:
                raise InvalidOperationException("Document must have a @metadata field")
            document_id = metadata.get('@id', None)
            if not id:
                raise InvalidOperationException("Document must have a @id field")
            last_change_vector = metadata.get('@change-vector', None)
            if not last_change_vector:
                raise InvalidOperationException("Document must have a @change-vector field")
            self._logger.info(
                "Got {0} (change vector: [{1}], size {2}".format(document_id, last_change_vector,
                                                                 len(item["Data"])))

            self.items.append(
                _SubscriptionBatchItem(change_vector=last_change_vector, document_id=document_id,
                                       raw_result=item["Data"], raw_metadata=metadata, result=entity,
                                       exception_message=item.get("Exception", None)))
        return last_change_vector



from abc import ABCMeta, abstractmethod
from pyravendb.tools.concurrentset import ConcurrentSet
from concurrent.futures import Future
from threading import Lock


class Observable:
    def __init__(self, on_connect=None, on_disconnect=None, executor=None):
        self.on_connect = on_connect
        self._on_disconnect = on_disconnect
        self.last_exception = None
        self.value_lock = Lock()
        self._value = 0
        self._future = None
        self._future_set = Future()
        self._subscribers = ConcurrentSet()
        self._executor = executor

    def __call__(self, filter_method):
        self._filter = filter_method
        return self

    def on_document_change_notification(self, value):
        try:
            if self._filter(value):
                for subscriber in self._subscribers:
                    subscriber.on_next(value)
        except Exception as e:
            self.error(e)

    def subscribe(self, observer):
        """
        @param Observer or func observer: The observer that will do action when changes happens
        :return: method that close the subscriber
        """
        self.inc()
        if not isinstance(observer, Observer):
            observer = ActionObserver(on_next=observer)
        self._subscribers.add(observer)

        def close_action():
            self.dec()
            subscriber = self._subscribers.pop(observer)
            if subscriber.on_complete:
                subscriber.on_complete()

        return close_action

    def inc(self):
        with self.value_lock:
            self._value += 1

    def dec(self):
        with self.value_lock:
            self._value -= 1
            if self._value == 0:
                self.set(self._executor.submit(self._on_disconnect))

    def set(self, future):
        if not self._future_set.done():
            def done_callback(f):
                try:
                    if f.cancelled():
                        self._future_set.cancel()
                    elif f.exception():
                        self._future_set.set_exception(f.exception)
                    else:
                        self._future_set.set_result(None)
                except Exception:
                    raise

            future.add_done_callback(done_callback)
        self._future = future

    def error(self, exception):
        future = Future()
        self.set(future)
        future.set_exception(exception)
        self.last_exception = exception
        for subscriber in self._subscribers:
            subscriber.on_error(exception)

    def close(self):
        future = Future()
        self.set(future)
        future.cancel()

        for subscriber in self._subscribers:
            subscriber.on_completed()

    def ensure_subscribe_now(self):
        self._future.result() if self._future else self._future_set.result()

    def send(self, msg):
        try:
            if self._filter(msg):
                for subscriber in self._subscribers:
                    subscriber.on_next(msg)
        except Exception as e:
            self.error(e)
            return


class Observer(metaclass=ABCMeta):
    @abstractmethod
    def on_completed(self):
        pass

    @abstractmethod
    def on_error(self, exception):
        pass

    @abstractmethod
    def on_next(self, value):
        pass


class ActionObserver(Observer):
    def __init__(self, on_next, on_error=None, on_completed=None):
        self._on_next = on_next
        self._on_error = on_error
        self._on_completed = on_completed

    def on_next(self, value):
        self._on_next(value)

    def on_error(self, exception):
        if self._on_error:
            self._on_error(exception)

    def on_completed(self):
        if self._on_completed:
            self._on_completed()



from pyravendb.tests.test_base import TestBase
from pyravendb.store.document_store import DocumentStore
from pyravendb.raven_operations.server_operations import CreateDatabaseOperation, DeleteDatabaseOperation
import hashlib
from shutil import rmtree
import os
import unittest

TOPOLOGY_FILES_DIR = os.path.join(os.getcwd(), "topology_files")
DATABASE = "SystemTest"


class Author:
    def __init__(self, name):
        self.name = name


class TestSystemTopologyCreation(TestBase):
    def tearDown(self):
        self.store.maintenance.server.send(
            DeleteDatabaseOperation(database_name=DATABASE, hard_delete=True))
        super(TestSystemTopologyCreation, self).tearDown()
        TestBase.delete_all_topology_files()
        if os.path.exists(TOPOLOGY_FILES_DIR):
            try:
                rmtree(TOPOLOGY_FILES_DIR, ignore_errors=True)
            except OSError as ex:
                print(ex)

    def test_topology_creation(self):
        created = False
        while not created:
            try:
                self.store.maintenance.server.send(CreateDatabaseOperation(database_name=DATABASE))
                created = True
            except Exception as e:
                if "already exists!" in str(e):
                    self.store.maintenance.server.send(
                        DeleteDatabaseOperation(database_name=DATABASE, hard_delete=True))
                    continue
                raise
        TestBase.wait_for_database_topology(self.store, DATABASE)

        with DocumentStore(urls=self.default_urls, database=DATABASE) as store:
            store.initialize()
            with store.open_session() as session:
                session.store(Author("Idan"))
                session.save_changes()
        topology_hash = hashlib.md5(
            "{0}{1}".format(self.default_urls[0], DATABASE).encode(
                'utf-8')).hexdigest()

        cluster_topology_hash = hashlib.md5(
            "{0}".format(self.default_urls[0]).encode(
                'utf-8')).hexdigest()

        self.assertTrue(os.path.exists(os.path.join(TOPOLOGY_FILES_DIR, topology_hash + ".raven-topology")))
        self.assertTrue(
            os.path.exists(os.path.join(TOPOLOGY_FILES_DIR, cluster_topology_hash + ".raven-cluster-topology")))


if __name__ == "__main__":
    unittest.main()



from pyravendb.tests.test_base import TestBase
from pyravendb.store.document_store import DocumentStore
from pyravendb.custom_exceptions import exceptions
import unittest


class Foo(object):
    def __init__(self, name, key):
        self.name = name
        self.key = key


class TestSessionStore(TestBase):
    def setUp(self):
        super(TestSessionStore, self).setUp()

    def tearDown(self):
        super(TestSessionStore, self).tearDown()
        self.delete_all_topology_files()

    def test_store_without_key(self):
        foo = Foo("test", 10)
        with self.store.open_session() as session:
            session.store(foo)
            session.save_changes()

        with self.store.open_session() as session:
            self.assertIsNotNone(session.load("foos/1-A"))

    def test_store_with_key(self):
        foo = Foo("test", 20)
        with self.store.open_session() as session:
            session.store(foo, "testingStore/1-A")
            session.save_changes()

        with self.store.open_session() as session:
            self.assertEqual(session.load("testingStore/1-A").key, 20)

    def test_store_after_delete_fail(self):
        foo = Foo("test", 20)
        with self.store.open_session() as session:
            session.store(foo, "testingStore")
            session.save_changes()

        with self.store.open_session() as session:
            session.delete("testingStore")
            with self.assertRaises(exceptions.InvalidOperationException):
                session.store(foo, "testingStore")

    def _save_documents(self, documents):
        with self.store.open_session() as session:
            for document in documents:
                session.store(document)
            session.save_changes()

    def test_store_the_same_documents_should_work(self):
        documents = []
        for i in range(0, 40):
            documents.append(Foo("IdanHaim", i))

        self._save_documents(documents)
        self._save_documents(documents)
        self._save_documents(documents)
        self._save_documents(documents)

        with self.store.open_session() as session:
            results = list(session.query().raw_query("From Foos"))
            self.assertEqual(len(results), 40 * 4)


if __name__ == "__main__":
    unittest.main()



from time import time as _time
import queue as queue


class IndexQueue(queue.Queue, object):
    def __init__(self):
        super(IndexQueue, self).__init__()

    def peek(self, index=0):
        if not self.qsize():
            return None
        return self.queue[index]

    def get(self, index=None, block=True, timeout=None):
        """Remove and return an item from the queue.

        If optional 'index' is not None the get will return the item in the index place
        If optional args 'block' is true and 'timeout' is None (the default),
        block if necessary until an item is available. If 'timeout' is
        a non-negative number, it blocks at most 'timeout' seconds and raises
        the Empty exception if no item was available within that time.
        Otherwise ('block' is false), return an item if one is immediately
        available, else raise the Empty exception ('timeout' is ignored
        in that case).
        """
        self.not_empty.acquire()
        try:
            if not block:
                if not self._qsize():
                    raise queue.Empty
            elif timeout is None:
                while not self._qsize():
                    self.not_empty.wait()
            elif timeout < 0:
                raise ValueError("'timeout' must be a non-negative number")
            else:
                endtime = _time() + timeout
                while not self._qsize():
                    remaining = endtime - _time()
                    if remaining <= 0.0:
                        raise queue.Empty
                    self.not_empty.wait(remaining)
            if not index or index == 0:
                item = self._get()
            else:
                item = self._get_with_index(index)
            self.not_full.notify()
            return item
        finally:
            self.not_empty.release()

    def _get_with_index(self, index):
        value = self.peek(index)
        self.queue.remove(value)
        return value

    def __lt__(self, other):
        return self.__key__() < other.__key__()

    def __len__(self):
        return self.queue.__len__()



from setuptools import setup, find_packages

setup(
    name='pyravendb',
    packages=find_packages(),
    version='4.0.4.9',
    long_description=open("README.rst").read(),
    description='This is the official python client for RavenDB v4.0 document database',
    author='RavenDB',
    author_email='support@ravendb.net',
    url='https://github.com/ravendb/RavenDB-Python-Client',
    license='MIT',
    keywords='pyravendb',
    install_requires=
    [
        'requests >= 2.18.4',
        'inflector >= 2.0.12',
        'xxhash >= 1.0.1',
        'pyOpenSSL >= 17.2.0',
        'ijson >= 2.3',
        'websocket-client >= 0.46.0'
    ],
    zip_safe=False
)



from pyravendb.tests.test_base import TestBase
from pyravendb.data.indexes import IndexDefinition
from pyravendb.raven_operations.maintenance_operations import PutIndexesOperation
from pyravendb.custom_exceptions.exceptions import InvalidOperationException
import unittest


class User:
    def __init__(self, name, age):
        self.name = name
        self.age = age


class TestAdvanced(TestBase):
    def tearDown(self):
        super(TestAdvanced, self).tearDown()
        self.delete_all_topology_files()

    def test_stream_query(self):
        maps = ("from user in docs.Users "
                "select new {"
                "name = user.name,"
                "age = user.age}")
        index_definition = IndexDefinition(name="UserByName", maps=maps)

        self.store.maintenance.send(PutIndexesOperation(index_definition))

        with self.store.open_session() as session:
            for i in range(0, 12000):
                session.store(User("Idan", i))
            session.save_changes()

        with self.store.open_session() as session:
            query = session.query(object_type=User, index_name="UserByName")
            results = session.advanced.stream(query)
            result_counter = 0
            for _ in results:
                result_counter += 1
            self.assertTrue(result_counter == 12000)

    def test_put_attachment(self):
        with self.store.open_session() as session:
            session.store(User("Idan", 30), "users/1-A")
            session.save_changes()

        with self.store.open_session() as session:
            with open("output.txt", "rb") as binary_list:
                user = session.load("users/1-A")
                session.advanced.attachment.store(user, "my_text_file", binary_list, content_type="text/plain")
                session.save_changes()

        with self.store.open_session() as session:
            attachment = session.advanced.attachment.get("users/1-A", "my_text_file")
            self.assertIsNotNone(attachment)

    def test_put_attachment_with_store(self):
        with self.store.open_session() as session:
            session.store(User("Idan", 30), "users/1-A")
            session.save_changes()

        with self.store.open_session() as session:
            with open("output.txt", "rb") as binary_list:
                session.store(User("Ilay", 4), "users/2-A")
                session.advanced.attachment.store("users/1-A", "my_text_file", binary_list, content_type="text/plain")
                session.save_changes()

        with self.store.open_session() as session:
            user = session.load("users/2-A")
            attachment = session.advanced.attachment.get("users/1-A", "my_text_file")
            self.assertIsNotNone(attachment)
            self.assertIsNotNone(user)
            self.assertTrue(user.name == "Ilay")

    def test_delete_attachment(self):
        with self.store.open_session() as session:
            session.store(User("Idan", 30), "users/1-A")
            session.save_changes()

        with self.store.open_session() as session:
            with open("output.txt", "rb") as binary_list:
                session.advanced.attachment.store("users/1-A", "my_text_file", binary_list, content_type="text/plain")
                session.save_changes()

        with self.store.open_session() as session:
            attachment = session.advanced.attachment.get("users/1-A", "my_text_file")
            self.assertIsNotNone(attachment)
            session.advanced.attachment.delete("users/1-A", "my_text_file")
            session.save_changes()
            attachment = session.advanced.attachment.get("users/1-A", "my_text_file")
            self.assertIsNone(attachment)

    def test_try_delete_attachment_putted_in_the_same_session(self):
        with self.store.open_session() as session:
            session.store(User("Idan", 30), "users/1-A")
            session.save_changes()

        with self.store.open_session() as session:
            with open("output.txt", "rb") as binary_list:
                session.advanced.attachment.store("users/1-A", "my_text_file", binary_list, content_type="text/plain")
                with self.assertRaises(InvalidOperationException):
                    session.advanced.attachment.delete("users/1-A", "my_text_file")


if __name__ == "__main__":
    unittest.main()



from pyravendb.commands.raven_commands import QueryCommand
from pyravendb.custom_exceptions.exceptions import *
import logging

logging.basicConfig(filename='query.log', level=logging.DEBUG)


class QueryOperation:
    def __init__(self, session, index_name, index_query, disable_entities_tracking=False, metadata_only=False,
                 index_entries_only=False):
        self._session = session
        self._index_name = index_name
        self._index_query = index_query
        self._disable_entities_tracking = disable_entities_tracking
        self._metadata_only = metadata_only
        self._index_entries_only = index_entries_only

        if self._session.conventions.raise_if_query_page_size_is_not_set and not self._index_query.page_size_set:
            raise InvalidOperationException("""Attempt to query without explicitly specifying a page size.
                                            You can use .take() methods to set maximum number of results.
                                            By default the page size is set to sys.maxsize and can cause
                                            severe performance degradation.""")

    @property
    def index_query(self):
        return self._index_query

    def create_request(self):
        self._session.increment_requests_count()
        logging.info("Executing query '{0}' on index '{1}'".format(self._index_query.query, self._index_name))

        return QueryCommand(self._session.conventions, self._index_query, self._metadata_only, self._index_entries_only)



import unittest
from pyravendb.commands.raven_commands import BatchCommand, GetDocumentCommand
from pyravendb.commands import commands_data
from pyravendb.data.patches import PatchRequest
from pyravendb.tests.test_base import TestBase


class TestDelete(TestBase):
    def setUp(self):
        super(TestDelete, self).setUp()
        self.put_command1 = commands_data.PutCommandData("products/1-A",
                                                         document={"Name": "tests", "Category": "testing"},
                                                         metadata={"Raven-Python-Type": "Products",
                                                                   "@collection": "Products"})
        self.put_command2 = commands_data.PutCommandData("products/2-A",
                                                         document={"Name": "testsDelete", "Category": "testing"},
                                                         metadata={"Raven-Python-Type": "Products",
                                                                   "@collection": "Products"})
        self.delete_command = commands_data.DeleteCommandData("products/2-A")
        patch = PatchRequest("this.Name = 'testing';")
        self.scripted_patch_command = commands_data.PatchCommandData("products/1-A", patch)
        self.requests_executor = self.store.get_request_executor()

    def tearDown(self):
        super(TestDelete, self).tearDown()
        self.delete_all_topology_files()

    def test_success_one_command(self):
        batch_command = BatchCommand([self.put_command1])
        batch_result = self.requests_executor.execute(batch_command)
        self.assertEqual(len(batch_result), 1)

    def test_success_multi_commands(self):
        batch_command = BatchCommand([self.put_command1, self.put_command2, self.delete_command])
        batch_result = self.requests_executor.execute(batch_command)
        self.assertEqual(len(batch_result), 3)

    def test_scripted_patch(self):
        batch_command = BatchCommand([self.put_command1, self.scripted_patch_command])
        self.requests_executor.execute(batch_command)
        result = self.requests_executor.execute(GetDocumentCommand("products/1-A"))
        self.assertEqual(result["Results"][0]["Name"], "testing")

    def test_fail(self):
        with self.assertRaises(ValueError):
            batch_command = BatchCommand([self.put_command1, self.put_command2, None])
            self.requests_executor.execute(batch_command)

    if __name__ == "__main__":
        unittest.main()



import unittest
from pyravendb.commands.raven_commands import *
from pyravendb.custom_exceptions import exceptions
from pyravendb.tests.test_base import TestBase


class TestDelete(TestBase):
    def setUp(self):
        super(TestDelete, self).setUp()
        self.requests_executor = self.store.get_request_executor()
        self.response = self.requests_executor.execute(
            PutDocumentCommand("products/101", {"Name": "test", "@metadata": {}}))

        self.requests_executor.execute(PutDocumentCommand("products/10", {"Name": "test", "@metadata": {}}))
        self.other_response = self.requests_executor.execute(
            PutDocumentCommand("products/102", {"Name": "test", "@metadata": {}}))

    def tearDown(self):
        super(TestDelete, self).tearDown()
        self.delete_all_topology_files()

    def test_delete_success_no_change_vector(self):
        delete_command = DeleteDocumentCommand("products/10")
        self.assertIsNone(self.requests_executor.execute(delete_command))

    def test_delete_success_with_change_vector(self):
        delete_command = DeleteDocumentCommand("products/102", self.other_response["ChangeVector"])
        self.assertIsNone(self.requests_executor.execute(delete_command))

    def test_delete_fail(self):
        with self.assertRaises(Exception):
            self.requests_executor.execute(DeleteDocumentCommand("products/101", self.response["ChangeVector"] + '10'))

    if __name__ == "__main__":
        unittest.main()



from threading import Lock
from pyravendb.changes.observers import Observable
from pyravendb.subscriptions.data import IncrementalJsonParser
import websocket
from pyravendb.custom_exceptions.exceptions import NotSupportedException
from pyravendb.custom_exceptions.exceptions import ChangeProcessingException
from pyravendb.tools.utils import Utils
import copy
from time import sleep
from concurrent.futures import ThreadPoolExecutor, TimeoutError, Future
import logging
import sys


class DatabaseChanges:
    def __init__(self, request_executor, database_name, on_close, on_error=None, executor=None):
        self._request_executor = request_executor
        self._conventions = request_executor.conventions
        self._database_name = database_name

        self._command_id = 0
        self.client_websocket = websocket.WebSocket()
        self._closed = False
        self._on_close = on_close
        self.on_error = on_error
        self._observables = dict()

        self._executor = executor if executor else ThreadPoolExecutor(max_workers=10)
        self._worker = self._executor.submit(self.do_work)
        self.send_lock = Lock()
        self._confirmations_lock = Lock()
        self._confirmations = {}

        self._command_id = 0
        self._immediate_connection = 0

        self._logger = logging.getLogger("database_changes")
        handler = logging.FileHandler('changes.log')
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        handler.setFormatter(formatter)
        self._logger.addHandler(handler)
        self._logger.setLevel(logging.DEBUG)

    def do_work(self):
        preferred_node = self._request_executor.get_preferred_node()
        url = "{0}/databases/{1}/changes".format(preferred_node.url, self._database_name).lower(). \
            replace("http://", "ws://").replace("https://", "wss://").replace(".fiddler", "")

        while not self._closed:
            try:
                if not self.client_websocket.connected:
                    if self._request_executor.certificate:
                        self.client_websocket.sock_opt.sslopt.update({'ca_certs': self._request_executor.certificate})
                    self.client_websocket.connect(url)
                    for observables in self._observables.values():
                        for observer in observables.values():
                            observer.set(self._executor.submit(observer.on_connect))
                    self._immediate_connection = 1
                self.process_changes()
            except ChangeProcessingException as e:
                self.notify_about_error(e)
                continue
            except websocket.WebSocketException as e:
                if getattr(sys, 'gettrace', None):
                    self._logger.error(e)
                self.notify_about_error(e)
                return
            except Exception as e:
                self._immediate_connection = 0
                self.client_websocket = websocket.WebSocket()
                self.notify_about_error(e)
            sleep(1)

    def process_changes(self):
        parser = IncrementalJsonParser(self.client_websocket, is_websocket=True)
        while not self._closed:
            try:
                response = parser.next_object()
                if response:
                    response_type = response.get("Type", None)
                    if not response_type:
                        continue
                    if response_type == "Error":
                        exception = response["Exception"]
                        self.notify_about_error(Exception(exception))
                    elif response_type == "Confirm":
                        command_id = response.get("CommandId", None)
                        if command_id and command_id in self._confirmations:
                            with self._confirmations_lock:
                                future = self._confirmations.pop(command_id)
                                future.set_result("done complete future")
                    else:
                        value = response.get("Value", None)
                        if value:
                            if response_type not in ("DocumentChange", "IndexChange", "OperationsStatusChange"):
                                raise NotSupportedException(response_type)
                            self._notify_subscribers(value, copy.copy(self._observables[response_type]))
            except Exception as e:
                self.notify_about_error(e)
                raise ChangeProcessingException(e)

    def _notify_subscribers(self, value, observables):
        for observable in observables.values():
            observable.send(value)

    def close(self):
        self._closed = True
        self.client_websocket.close()

        for observable in self._observables.values():
            for observer in observable.values():
                observer.close()

        with self._confirmations_lock:
            for confirmation in self._confirmations.values():
                confirmation.cancel()

        self._observables.clear()
        if self._on_close:
            self._on_close(self._database_name)

        self._executor.shutdown(wait=True)

    def notify_about_error(self, e):
        if self.on_error:
            self.on_error(e)

        for observables in self._observables.values():
            for observer in observables.values():
                observer.error(e)

        for _, observable in self._observables.items():
            observable.error(e)

    def for_all_documents(self):
        observable = self.get_or_add_observable("DocumentChange", "all-docs", "watch-docs", "unwatch-docs", None)(
            lambda x: True)
        return observable

    def for_all_operations(self):
        observable = self.get_or_add_observable("OperationsStatusChange", "all-operations", "watch-operations",
                                                "unwatch-operations", None)(lambda x: True)
        return observable

    def for_all_indexes(self):
        observable = self.get_or_add_observable("IndexChange", "all-indexes", "watch-indexes", "unwatch-indexes",
                                                None)(lambda x: True)
        return observable

    def for_index(self, index_name):
        observable = self.get_or_add_observable("IndexChange", "indexes/" + index_name, "watch-index", "unwatch-index",
                                                index_name)(lambda x: x["Name"].casefold() == index_name.casefold())
        return observable

    def for_operation_id(self, operation_id):
        observable = self.get_or_add_observable("OperationsStatusChange", "operations/" + operation_id,
                                                "watch-operation", "unwatch-operation", str(operation_id))(
            lambda x: x["operationId"] == str(operation_id))
        return observable

    def for_document(self, doc_id):
        observable = self.get_or_add_observable("DocumentChange", "docs/" + doc_id, "watch-doc", "unwatch-doc", doc_id)(
            lambda x: x["Id"].casefold() == doc_id.casefold())
        return observable

    def for_documents_start_with(self, doc_id_prefix):
        observable = self.get_or_add_observable("DocumentChange", "prefixes/" + doc_id_prefix, "watch-prefix",
                                                "unwatch-prefix", doc_id_prefix)(
            lambda x: x['Id'] is not None and x["Id"].casefold().startswith(doc_id_prefix.casefold()))
        return observable

    def for_documents_in_collection(self, collection_name):
        observable = self.get_or_add_observable("DocumentChange", "collections/" + collection_name, "watch-collection",
                                                "unwatch-collection", collection_name)(
            lambda x: x["CollectionName"].casefold().starts_with(collection_name.casefold()))
        return observable

    def get_or_add_observable(self, group, name, watch_command, unwatch_command, value):
        if group not in self._observables:
            self._observables[group] = {}

        if name not in self._observables[group]:
            def on_disconnect():
                try:
                    if self.client_websocket.connected:
                        self.send(unwatch_command, value)
                except websocket.WebSocketException:
                    pass

            def on_connect():
                self.send(watch_command, value)

            observable = Observable(on_connect=on_connect, on_disconnect=on_disconnect, executor=self._executor)
            self._observables[group][name] = observable
            if self._immediate_connection != 0:
                observable.set(self._executor.submit(observable.on_connect))
        return self._observables[group][name]

    def send(self, command, value):
        current_command_id = 0
        future = Future()
        try:
            with self.send_lock:
                self._command_id += 1
                current_command_id += self._command_id
                data = Utils.dict_to_bytes({"CommandId": current_command_id, "Command": command, "Param": value})
                with self._confirmations_lock:
                    self._confirmations[current_command_id] = future
                self.client_websocket.send(data)

                try:
                    future.result(timeout=15)
                except TimeoutError:
                    future.cancel()
                    raise TimeoutError("Did not get a confirmation for command #" + current_command_id)
                except Exception as e:
                    if getattr(sys, 'gettrace', None):
                        self._logger.info('The coroutine raised an exception: {!r}'.format(e))
        except websocket.WebSocketConnectionClosedException:
            pass



from pyravendb.commands.raven_commands import RavenCommand, GetDatabaseRecordCommand
from pyravendb.data.certificate import CertificateDefinition, SecurityClearance
from pyravendb.custom_exceptions import exceptions
from requests.exceptions import RequestException
from pyravendb.tools.utils import Utils
from abc import abstractmethod


class ServerOperation(object):
    __slots__ = ['__operation']

    def __init__(self):
        self.__operation = "ServerOperation"

    @property
    def operation(self):
        return self.__operation

    @abstractmethod
    def get_command(self, conventions):
        raise NotImplementedError


class CreateDatabaseOperation(ServerOperation):
    def __init__(self, database_name, replication_factor=1, settings=None, secure_settings=None):
        super(CreateDatabaseOperation, self).__init__()
        self.replication_factor = replication_factor
        self.database_record = CreateDatabaseOperation.get_default_database_record()
        Utils.database_name_validation(database_name)
        self.database_record["DatabaseName"] = database_name
        if settings:
            self.database_record["Settings"] = settings
        if secure_settings:
            self.database_record["SecuredSettings"] = secure_settings

    def get_command(self, conventions):
        return self._CreateDatabaseCommand(self.database_record, self.replication_factor, conventions)

    @staticmethod
    def get_default_database_record():
        return {"DatabaseName": None, "Disabled": False, "Encrypted": False, "DeletionInProgress": None,
                "DataDirectory": None, "Topology": None,
                "ConflictSolverConfig": {"DatabaseResolverId": None, "ResolveByCollection": None,
                                         "ResolveToLatest": False},
                "Indexes": None, "AutoIndexes": None, "Transformers": None, "Identities": None, "Settings": {},
                "ExternalReplication": [], "RavenConnectionStrings": {}, "SqlConnectionStrings": {}, "RavenEtls": None,
                "SqlEtls": None, "SecuredSettings": None, "Revisions": None, "Expiration": None,
                "PeriodicBackups": None, "Client": None, "CustomFunctions": None}

    class _CreateDatabaseCommand(RavenCommand):
        def __init__(self, database_record, replication_factor, convention):
            if convention is None:
                raise ValueError("Invalid convention")
            super(CreateDatabaseOperation._CreateDatabaseCommand, self).__init__(method="PUT")

            self._database_record = database_record
            self._replication_factor = replication_factor
            self.convention = convention

        def create_request(self, server_node):
            self.url = "{0}/admin/databases?name={1}".format(server_node.url, self._database_record["DatabaseName"])
            self.url += "&replication-factor=" + str(self._replication_factor)

            self.data = self._database_record

        def set_response(self, response):
            if response is None:
                raise ValueError("response is invalid.")

            if response.status_code == 201:
                return response.json()

            if response.status_code == 400:
                response = response.json()
                if "Error" in response:
                    raise RequestException(response["Error"])


class DeleteDatabaseOperation(ServerOperation):
    def __init__(self, database_name, hard_delete, from_node=None, time_to_wait_for_confirmation=None):
        if database_name is None:
            raise ValueError("Invalid database_name")

        super(DeleteDatabaseOperation, self).__init__()
        self._parameters = {"DatabaseNames": [database_name], "HardDelete": hard_delete,
                            "TimeToWaitForConfirmation": time_to_wait_for_confirmation}
        if from_node:
            self._parameters["FromNodes"] = [from_node]

    def get_command(self, conventions):
        return self._DeleteDatabaseCommand(self._parameters)

    class _DeleteDatabaseCommand(RavenCommand):
        def __init__(self, parameters):
            if parameters is None:
                raise ValueError("Invalid parameters")

            self._parameters = parameters

            super(DeleteDatabaseOperation._DeleteDatabaseCommand, self).__init__(method="DELETE")

        def create_request(self, server_node):
            self.url = "{0}/admin/databases".format(server_node.url)
            self.data = self._parameters

        def set_response(self, response):
            try:
                response = response.json()
                if "Error" in response:
                    raise exceptions.DatabaseDoesNotExistException(response["Message"])
            except ValueError:
                raise response.raise_for_status()
            return {"raft_command_index": response["RaftCommandIndex"]}


class GetDatabaseNamesOperation(ServerOperation):
    def __init__(self, start, page_size):
        super(GetDatabaseNamesOperation, self).__init__()
        self._start = start
        self._page_size = page_size

    def get_command(self, conventions):
        return self._GetDatabaseNamesCommand(self._start, self._page_size)

    class _GetDatabaseNamesCommand(RavenCommand):
        def __init__(self, start, page_size):
            super(GetDatabaseNamesOperation._GetDatabaseNamesCommand, self).__init__(method="GET", is_read_request=True)
            self._start = start
            self._page_size = page_size

        def create_request(self, server_node):
            self.url = "{0}/databases?start={1}&pageSize={2}&namesOnly=true".format(server_node.url, self._start,
                                                                                    self._page_size)

        def set_response(self, response):
            if response is None:
                raise ValueError("Invalid response")

            response = response.json()
            if "Error" in response:
                raise exceptions.ErrorResponseException(response["Error"])

            if "Databases" not in response:
                raise ValueError("Invalid response")

            return response["Databases"]


class GetDatabaseRecordOperation(ServerOperation):
    def __init__(self, database_name):
        super(GetDatabaseRecordOperation, self).__init__()
        self._database_name = database_name

    def get_command(self, conventions):
        return GetDatabaseRecordCommand(self._database_name)


# --------------------CertificateOperation------------------------------

class GetCertificateOperation(ServerOperation):
    def __init__(self, start, page_size):

        super(GetCertificateOperation, self).__init__()
        self._start = start
        self._page_size = page_size

    def get_command(self, conventions):
        return self._GetCertificateCommand(self._start, self._page_size)

    class _GetCertificateCommand(RavenCommand):
        def __init__(self, start, page_size):
            super(GetCertificateOperation._GetCertificateCommand, self).__init__(method="GET")
            self._start = start
            self._page_size = page_size

        def create_request(self, server_node):
            self.url = "{0}/admin/certificates?start={1}&pageSize={2}".format(server_node.url, self._start,
                                                                              self._page_size)

        def set_response(self, response):
            if response is None:
                raise ValueError("response is invalid.")
            data = {}
            try:
                response = response.json()["Results"]
                if len(response) > 1:
                    raise ValueError("response is Invalid")
                for key, value in response[0].items():
                    data[Utils.convert_to_snake_case(key)] = value
                return CertificateDefinition(**data)

            except ValueError:
                raise response.raise_for_status()


class CreateClientCertificateOperation(ServerOperation):
    def __init__(self, name, permissions, clearance=SecurityClearance.unauthenticated_clients, password=None):
        """
        Add certificate json to the server and get certificate from server to use
        @param str name: The name of the certificate
        @param Dict[str:DatabaseAccess] permissions: the permissions to the database the key is the name of the database
        the value is database access (read or admin)
        @param SecurityClearance clearance: The clearance of the client
        @param str password: The password of the certificate
        """
        if name is None:
            raise ValueError("name cannot be None")
        if permissions is None:
            raise ValueError("permissions cannot be None")

        super(CreateClientCertificateOperation, self).__init__()
        self._name = name
        self._permissions = permissions
        self._clearance = clearance
        self._password = password

    def get_command(self, conventions):
        return self._CreateClientCertificateCommand(self._name, self._permissions, self._clearance, self._password)

    class _CreateClientCertificateCommand(RavenCommand):
        def __init__(self, name, permissions, clearance, password):
            if name is None:
                raise ValueError("name cannot be None")
            if permissions is None:
                raise ValueError("permissions cannot be None")

            super(CreateClientCertificateOperation._CreateClientCertificateCommand, self).__init__(method="POST",
                                                                                                   is_read_request=True,
                                                                                                   use_stream=True)
            self._name = name
            self._permissions = permissions
            self._clearance = clearance
            self._password = password

        def create_request(self, server_node):
            self.url = server_node.url + "/admin/certificates"
            self.data = {"Name": self._name, "SecurityClearance": str(self._clearance)}
            if self._password:
                self.data["Password"] = self._password

            permissions = {}
            for key, value in self._permissions.items():
                permissions.update({"Database": key, "Access": str(self.value)})

            self.data["Permissions"] = permissions

        def set_response(self, response):
            if response is None:
                return None

            if response.status_code != 201:
                response = response.json()
                if "Error" in response:
                    raise exceptions.ErrorResponseException(response["Error"])
            return response.raw.data


class PutClientCertificateOperation(ServerOperation):
    def __init__(self, name, certificate, permissions, clearance=SecurityClearance.unauthenticated_clients):
        """
        @param str name: Certificate name
        @param x509 certificate: X509 certificate file (OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, pem))
        @param Dict[str:DatabaseAccess] permissions: the permissions to the client
        @param SecurityClearance clearance: The clearance of the client
        """
        if certificate is None:
            raise ValueError("certificate cannot be None")
        if permissions is None:
            raise ValueError("permissions cannot be None")

        super(PutClientCertificateOperation, self).__init__()
        self._name = name
        self._certificate = certificate
        self._permissions = permissions
        self._clearance = clearance

    def get_command(self, conventions):
        return self._PutClientCertificateCommand(self._name, self._certificate, self._permissions, self._clearance)

    class _PutClientCertificateCommand(RavenCommand):
        def __init__(self, name, certificate, permissions, clearance=SecurityClearance.unauthenticated_clients):
            if certificate is None:
                raise ValueError("certificate cannot be None")
            if permissions is None:
                raise ValueError("permissions cannot be None")

            super(PutClientCertificateOperation._PutClientCertificateCommand, self).__init__(method="PUT",
                                                                                             use_stream=True)
            self._name = name
            self._certificate = certificate
            self._permissions = permissions
            self._clearance = clearance

        def create_request(self, server_node):
            self.url = server_node.url + "/admin/certificates"

            self.data = {"Name": self._name, "Certificate": self._certificate,
                         "SecurityClearance": str(self._clearance)}
            permissions = []
            for key, value in self._permissions.items():
                permissions.append({"Database": key, "Access": str(self.value)})

            self.data["Permissions"] = permissions

        def set_response(self, response):
            pass


class DeleteCertificateOperation(ServerOperation):
    def __init__(self, thumbprint):
        if thumbprint is None:
            raise ValueError("certificate cannot be None")
        super(DeleteCertificateOperation, self).__init__()
        self._thumbprint = thumbprint

    def get_command(self, conventions):
        return self._DeleteCertificateCommand(self._thumbprint)

    class _DeleteCertificateCommand(RavenCommand):
        def __init__(self, thumbprint):
            if thumbprint is None:
                raise ValueError("certificate cannot be None")
            super(DeleteCertificateOperation._DeleteCertificateCommand, self).__init__(method="DELETE")
            self._thumbprint = thumbprint

        def create_request(self, server_node):
            self.url = "{0}/admin/certificates?thumbprint={1}".format(server_node.url,
                                                                      Utils.quote_key(self._thumbprint))

        def set_response(self, response):
            pass



from pyravendb.data.query import IndexQuery
from pyravendb.tests.test_base import TestBase
from pyravendb.commands.raven_commands import *
from pyravendb.custom_exceptions.exceptions import ErrorResponseException
from pyravendb.data.document_conventions import DocumentConventions
import time


class TestQuery(TestBase):
    def setUp(self):
        super(TestQuery, self).setUp()
        self.requests_executor = self.store.get_request_executor()
        self.requests_executor.execute(
            PutDocumentCommand("products/10", {"Name": "test", "@metadata": {"@collection": "Products",
                                                                             "Raven-Python-Type": "object"}}))

    def tearDown(self):
        super(TestQuery, self).tearDown()
        self.delete_all_topology_files()

    def test_only_query(self):
        query_command = QueryCommand(
            index_query=IndexQuery(query="From INDEX 'AllDocuments' WHERE Tag='Products'",
                                   wait_for_non_stale_results=True),
            conventions=DocumentConventions())
        response = self.requests_executor.execute(query_command)
        self.assertEqual(response["Results"][0]["Name"], "test")

    def test_get_only_metadata(self):
        query_command = QueryCommand(
            index_query=IndexQuery(query="From INDEX 'AllDocuments' WHERE Tag='Products'",
                                   wait_for_non_stale_results=True), conventions=DocumentConventions(),
            metadata_only=True)
        response = self.requests_executor.execute(query_command)
        self.assertFalse("Name" in response["Results"][0])

    def test_get_only_index_entries(self):
        query_command = QueryCommand(index_query=IndexQuery(query="FROM INDEX 'AllDocuments' WHERE Tag='products'",
                                                            wait_for_non_stale_results=True),
                                     conventions=DocumentConventions(),
                                     index_entries_only=True)

        start = time.time()
        while True and time.time() - start < 30:
            response = self.requests_executor.execute(query_command)
            if not response["IsStale"]:
                break
        self.assertFalse("@metadata" in response["Results"][0])

    def test_fail_with_no_index(self):
        query_command = QueryCommand(index_query=IndexQuery(query="From INDEX IndexIsNotExists WHERE Tag='products'",
                                                            wait_for_non_stale_results=True),
                                     conventions=DocumentConventions())
        self.assertIsNone(self.requests_executor.execute(query_command))

    def test_fail_with_wrong_query(self):
        with self.assertRaises(ErrorResponseException):
            query_command = QueryCommand(index_query=IndexQuery(query="From INDEX IndexIsNotExists WHERE Tag=products",
                                                                wait_for_non_stale_results=True),
                                         conventions=DocumentConventions())
            self.assertIsNone(self.requests_executor.execute(query_command))



import unittest
from pyravendb.commands.raven_commands import PutDocumentCommand, PatchCommand
from pyravendb.custom_exceptions.exceptions import DocumentDoesNotExistsException
from pyravendb.data.patches import PatchRequest
from pyravendb.tests.test_base import TestBase


class TestPatch(TestBase):
    def setUp(self):
        super(TestPatch, self).setUp()
        command = PutDocumentCommand("products/10", {"Name": "test", "@metadata": {}})
        self.requests_executor = self.store.get_request_executor()
        self.requests_executor.execute(command)

    def tearDown(self):
        super(TestPatch, self).tearDown()
        TestBase.delete_all_topology_files()

    def test_patch_success_ignore_missing(self):
        command = PatchCommand(document_id="products/10", change_vector=None,
                               patch=PatchRequest("this.Name = 'testing'"), patch_if_missing=None)
        result = self.requests_executor.execute(command)
        self.assertIsNotNone(result)
        self.assertTrue(result["ModifiedDocument"]["Name"] == "testing")

    def test_patch_success_not_ignore_missing(self):
        command = PatchCommand(document_id="products/10", change_vector=None,
                               patch=PatchRequest("this.Name = 'testing'"),
                               patch_if_missing=None, skip_patch_if_change_vector_mismatch=True)
        result = self.requests_executor.execute(command)
        self.assertIsNotNone(result)
        self.assertTrue(result["ModifiedDocument"]["Name"] == "testing")

    def test_patch_fail_not_ignore_missing(self):
        command = PatchCommand(document_id="products/110", change_vector=None,
                               patch=PatchRequest("this.Name = 'testing'"),
                               patch_if_missing=None)
        self.assertIsNone(self.requests_executor.execute(command))


if __name__ == "__main__":
    unittest.main
